from typing import *
import numpy as np

from luma.interface.util import Matrix
from luma.interface.exception import NotFittedError, UnsupportedParameterError
from luma.interface.super import Estimator, Evaluator, Supervised
from luma.metric.regression import MeanSquaredError


__all__ = (
    'SVR', 
    'KernelSVR'
)


class SVR(Estimator, Supervised):
    
    """
    Support Vector Regressor (SVR) is a supervised machine learning 
    algorithm used for regression tasks. It operates by determining a 
    hyperplane that best fits the input data, aiming to minimize the 
    difference between the predicted and actual values.
    
    Parameters
    ----------
    `C` : Regularization parameter
    `epsilon` : Epsilon-tube in the training loss function
    `batch_size` : Size of a single batch
    `learning_rate` : Step-size of gradient descent update
    `max_iter` : Number of iteration
    
    """
    
    def __init__(self, 
                 C: float = 1.0, 
                 epsilon: float = 0.1,
                 batch_size: int = 100, 
                 learning_rate: float = 0.001, 
                 max_iter: int = 1000, 
                 verbose: bool = False) -> None:
        self.C = C
        self.epsilon = epsilon
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.max_iter = max_iter
        self.verbose = verbose
        self._fitted = False

    def fit(self, X: Matrix, y: Matrix) -> 'SVR':
        m, n = X.shape
        id = np.arange(m)
        np.random.shuffle(id)

        self.bias = 0.0
        self.weight = np.zeros(n)
        
        for i in range(self.max_iter):
            for batch_point in range(0, m, self.batch_size):
                gradient_w, gradient_b = 0, 0
                for j in range(batch_point, batch_point + self.batch_size):
                    if j >= m: continue
                    idx = id[j]
                    pred = np.dot(self.weight, X[idx].T) + self.bias
                    loss = pred - y[idx]
                    
                    if np.abs(loss) > self.epsilon:
                        gradient_w += self.C * np.sign(loss) * X[idx]
                        gradient_b += self.C * np.sign(loss)

                self.weight -= self.learning_rate * self.weight
                self.weight -= self.learning_rate * gradient_w
                self.bias -= self.learning_rate * gradient_b

            if self.verbose and i % 100 == 0 and i:
                print(f'[SVR] Finished iteration {i}/{self.max_iter}', end=' ')
                print(f'with weight-norm of {np.linalg.norm(self.weight)}')

        self._fitted = True
        return self
    
    def predict(self, X: Matrix) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        return np.dot(X, self.weight) + self.bias

    def score(self, X: Matrix, y: Matrix, 
              metric: Evaluator = MeanSquaredError) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)

    def set_params(self,
                   C: float = None,
                   epsilon: float = None,
                   batch_size: int = None,
                   learning_rate: float = None,
                   max_iter: int = None) -> None:
        if C is not None: self.C = float(C)
        if epsilon is not None: self.epsilon = float(epsilon)
        if batch_size is not None: self.batch_size = int(batch_size)
        if learning_rate is not None: self.learning_rate = float(learning_rate)
        if max_iter is not None: self.max_iter = int(max_iter)


class KernelSVR(Estimator, Supervised):
    
    """
    Kernel Support Vector Regression (KernelSVR) is an extension of the 
    Support Vector Regression (SVR) algorithm that enables regression on 
    non-linearly related data by mapping it into a higher-dimensional space
    using a kernel function. The kernel function allows the algorithm to 
    establish a flexible and non-linear regression model in the transformed space.

    Parameters
    ----------
    `C` : Regularization parameter
    `deg` : Polynomial Degree for `poly` kernel
    `gamma` : Shape parameter of Gaussian curve for `rbf` kernel
    `coef` : Coefficient for `poly`, `sigmoid` kernel
    `learning_rate` : Step-size for gradient descent update
    `max_iter` : Number of iteration
    `kernel` : Type of kernel (e.g., `linear`, `poly`, `rbf`, `sigmoid`)

    """
    
    def __init__(self,
                 C: float = 1.0,
                 deg: int = 3,
                 gamma: float = 1.0,
                 coef: float = 1.0,
                 learning_rate: float = 0.001,
                 max_iter: int = 1000,
                 kernel: Literal['linear', 'poly', 'rbf', 'sigmoid'] = 'rbf',
                 verbose: bool = False) -> None:
        self.C = C
        self.deg = deg
        self.gamma = gamma
        self.coef = coef
        self.learning_rate = learning_rate
        self.max_iter = max_iter
        self.kernel = kernel
        self.verbose = verbose
        self._kernel_func = None
        self._fitted = False

    def fit(self, X: Matrix, y: Matrix) -> 'KernelSVR':
        m, _ = X.shape
        self._set_kernel_func()
        self.alpha = np.random.random(m)
        self.bias = 0
        
        y_mul_kernel = np.outer(y, y) * self._kernel_func(X, X)
        for i in range(self.max_iter):
            gradient = np.ones(m) - y_mul_kernel.dot(self.alpha)
            self.alpha += self.learning_rate * gradient
            self.alpha = np.clip(self.alpha, 0, self.C)
            
            if self.verbose and i % 100 == 0 and i:
                print(f'[KernelSVR] Finished iteration {i}/{self.max_iter}', end=' ')
                print(f'with alpha-norm: {np.linalg.norm(self.alpha)}')
        
        alpha_idx = np.where((self.alpha > 0) & (self.alpha < self.C))[0]
        bias_list = []
        for idx in alpha_idx:
            _append = y[idx] - (self.alpha * y).dot(self._kernel_func(X, X[idx]))
            bias_list.append(_append)
        self.bias = np.mean(bias_list)
        
        self._X = X
        self._y = y
        self._fitted = True
        return self
    
    def _linear_kernel(self, xi: Matrix, xj: Matrix) -> Matrix:
        return xi.dot(xj.T)

    def _poly_kernel(self, xi: Matrix, xj: Matrix) -> Matrix:
        return (self.coef + xi.dot(xj.T)) ** self.deg

    def _rbf_kernel(self, xi: Matrix, xj: Matrix) -> Matrix:
        norm = np.linalg.norm(xi[:, np.newaxis] - xj[np.newaxis, :], axis=2)
        return np.exp(-self.gamma * norm ** 2)

    def _sigmoid_kernel(self, xi: Matrix, xj: Matrix) -> Matrix:
        return np.tanh(2 * xi.dot(xj.T) + self.coef)

    def _set_kernel_func(self) -> None:
        if self.kernel == 'linear': self._kernel_func = self._linear_kernel
        elif self.kernel == 'poly': self._kernel_func = self._poly_kernel
        elif self.kernel == 'rbf': self._kernel_func = self._rbf_kernel
        elif self.kernel == 'sigmoid': self._kernel_func = self._sigmoid_kernel
        else: raise UnsupportedParameterError(self.kernel)

    def predict(self, X: Matrix) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        return (self.alpha * self._y).dot(self._kernel_func(self._X, X)) + self.bias
    
    def score(self, X: Matrix, y: Matrix, 
              metric: Evaluator = MeanSquaredError) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)
    
    def set_params(self,
                   C: float = None,
                   deg: int = None,
                   gamma: float = None,
                   coef: float = None,
                   learning_rate: float = None,
                   max_iter: int = None,
                   kernel: Literal = None) -> None:
        if C is not None: self.C = float(C)
        if deg is not None: self.deg = int(deg)
        if gamma is not None: self.gamma = float(gamma)
        if coef is not None: self.coef = float(coef)
        if learning_rate is not None: self.learning_rate = float(learning_rate)
        if max_iter is not None: self.max_iter = int(max_iter)
        if kernel is not None: self.kernel = str(kernel)




from typing import *
from scipy.special import psi
import numpy as np

from luma.interface.util import Matrix
from luma.interface.exception import NotFittedError, UnsupportedParameterError
from luma.interface.super import Estimator, Evaluator, Supervised
from luma.metric.regression import MeanSquaredError


__all__ = (
    'PoissonRegressor', 
    'NegativeBinomialRegressor', 
    'GammaRegressor',
    'BetaRegressor', 
    'InverseGaussianRegressor'
)


class PoissonRegressor(Estimator, Supervised):
    
    """
    Poisson regression is a model for predicting count or frequency data. 
    It models the relationship between features and counts, with a focus 
    on count data that follows a Poisson distribution. The model uses 
    the natural logarithm of expected counts as a link function and is 
    trained to minimize the negative log-likelihood.
    
    Parameters
    ----------
    `learning_rate` : Step size of the gradient descent update
    `max_iter` : Number of iteration
    `rho` : Balancing parameter of `elastic-net` 
    `alpha` : Regularization strength
    `regularization` : Regularization type (e.g. `l1`, `l2`, `elastic-net`)
    
    """
    
    def __init__(self,
                 learning_rate: float = 0.01,
                 max_iter: int = 100,
                 rho: float = 0.5,
                 alpha: float = 0.01,
                 regularization: Literal['l1', 'l2','elastic-net'] = None,
                 verbose: bool = False) -> None:
        self.learning_rate = learning_rate
        self.max_iter = max_iter
        self.regularization = regularization
        self.rho = rho
        self.alpha = alpha
        self.verbose = verbose
        self._fitted = False
    
    def link_funciton(self, X: Matrix) -> Matrix:
        return np.exp(np.dot(X, self.weights))
    
    def fit(self, X: Matrix, y: Matrix) -> 'PoissonRegressor':
        X = np.column_stack((np.ones(X.shape[0]), X))
        m, n = X.shape
        self.weights = np.zeros(n)
        
        for i in range(self.max_iter):
            predictions = self.link_funciton(X)
            gradient = np.dot(X.T, y - predictions) * self.learning_rate
            gradient /= m
            
            if self.regularization:
                regularization_term = self._regularization_term()
                gradient -= self.alpha * regularization_term / m
            
            if self.verbose and i % 10 == 0:
                print(f'[PoissonReg] iteration: {i}/{self.max_iter}', end='')
                print(f' - gradient-norm: {np.linalg.norm(gradient)}')
            
            self.weights += gradient
        self._fitted = True
        return self
    
    def _regularization_term(self) -> Matrix:
        if self.regularization == 'l1': return np.sign(self.weights)
        elif self.regularization == 'l2': return self.weights
        elif self.regularization == 'elastic-net':
            l1_term = np.sign(self.weights)
            l2_term = 2 * self.weights
            return (1 - self.rho) * l2_term + self.rho * l1_term
        elif self.regularization is None: return np.zeros_like(self.weights)
        else: raise UnsupportedParameterError(self.regularization)
    
    def predict(self, X: Matrix) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        X = np.column_stack((np.ones(X.shape[0]), X))
        predictions = self.link_funciton(X)
        return predictions
    
    def score(self, X: Matrix, y: Matrix, 
              metric: Evaluator = MeanSquaredError) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)

    def set_params(self,
                   learning_rate: float = None,
                   max_iter: int = None,
                   alpha: float = None,
                   rho: float = None,
                   regularization: Literal = None) -> None:
        if learning_rate is not None: self.learning_rate = float(learning_rate)
        if max_iter is not None: self.max_iter = int(max_iter)
        if alpha is not None: self.alpha = float(alpha)
        if rho is not None: self.rho = float(rho)
        if regularization is not None: self.regularization = str(regularization)


class NegativeBinomialRegressor(Estimator, Supervised):
    
    """
    Negative Binomial Regression is a statistical method used to 
    model count data, like the number of events or occurrences, 
    when the data shows overdispersion (variance is greater than mean) 
    compared to the Poisson distribution. It estimates how predictors 
    influence the count while allowing for flexible variance modeling.
    
    Parameters
    ----------
    `learning_rate` : Step size of the gradient descent update
    `max_iter` : Number of iteration
    `alpha` : Regularization strength
    `rho` : Balancing parameter of `elastic-net` 
    `phi` : Dispersion parameter
    `regularization` : Regularization type (e.g. `l1`, `l2`, `elastic-net`)
    
    """
    
    def __init__(self,
                 learning_rate: float = 0.01,
                 max_iter: int = 100,
                 alpha: float = 0.01,
                 rho: float = 0.5,
                 phi: float = 1.0,
                 regularization: Literal['l1', 'l2', 'elastic-net'] = None,
                 verbose: bool = False) -> None:
        self.learning_rate = learning_rate
        self.max_iter = max_iter
        self.alpha = alpha
        self.rho = rho
        self.phi = phi
        self.regularization = regularization
        self.verbose = verbose
        self._fitted = False
    
    def link_function(self, X: Matrix) -> Matrix:
        return np.log(1 + np.exp(np.dot(X, self.weights)))
    
    def fit(self, X: Matrix, y: Matrix) -> 'NegativeBinomialRegressor':
        X = np.column_stack((np.ones(X.shape[0]),  X))
        m, n = X.shape
        self.weights = np.zeros(n)
        
        for i in range(self.max_iter):
            mu = self.link_function(X)
            gradient = (y - mu) / (self.phi * (1 + mu / self.phi))
            gradient = np.dot(X.T, gradient) * self.learning_rate
            gradient /= m
            
            if self.regularization:
                regularization_term = self._regularization_term()
                gradient -= self.alpha * regularization_term / m
            
            if self.verbose and i % 10 == 0:
                print(f'[NegBinReg] iteration: {i}/{self.max_iter}', end='')
                print(f' - gradient-norm: {np.linalg.norm(gradient)}')
            
            self.weights += gradient
        self._fitted = True
        return self
    
    def _regularization_term(self) -> Matrix:
        if self.regularization == 'l1': return np.sign(self.weights)
        elif self.regularization == 'l2': return self.weights
        elif self.regularization == 'elastic-net':
            l1_term = np.sign(self.weights)
            l2_term = 2 * self.weights
            return (1 - self.rho) * l2_term + self.rho * l1_term
        elif self.regularization is None: return np.zeros_like(self.weights)
        else: raise UnsupportedParameterError(self.regularization)
    
    def predict(self, X: Matrix) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        X = np.column_stack((np.ones(X.shape[0]), X))
        mu = self.link_function(X)
        return mu
    
    def score(self, X: Matrix, y: Matrix, 
              metric: Evaluator = MeanSquaredError) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)
    
    def set_params(self,
                   learning_rate: float = None,
                   max_iter: int = None,
                   alpha: float = None,
                   rho: float = None,
                   phi: float = None,
                   regularization: Literal = None) -> None:
        if learning_rate is not None: self.learning_rate = float(learning_rate)
        if max_iter is not None: self.max_iter = int(max_iter)
        if alpha is not None: self.alpha = float(alpha)
        if rho is not None: self.rho = float(rho)
        if phi is not None: self.phi = float(phi)
        if regularization is not None: self.regularization = str(regularization)


class GammaRegressor(Estimator, Supervised):
    
    """
    Gamma regression is a statistical method for analyzing 
    data with skewed, positively skewed distributions and where 
    the variance increases with the mean. It models the 
    relationship between variables using the gamma distribution, 
    making it suitable for situations where traditional 
    linear regression doesn't apply.
    
    Parameters
    ----------
    `alpha` : Shape parameter of gamma distribution
    `beta` : Scale parameter of gamma distribution
    `learning_rate` : Step size of the gradient descent update
    `max_iter` : Number of iteration
    `reg_strength` : Regularization strength
    `rho` : Balancing parameter of `elastic-net` 
    `regularization` : Regularization type (e.g. `l1`, `l2`, `elastic-net`)
    
    """
    
    def __init__(self,
                 alpha: float = 1.0,
                 beta: float = 1.0,
                 learning_rate: float = 0.01,
                 max_iter: int = 100,
                 reg_strength: float = 0.01,
                 rho: float = 0.5,
                 regularization: Literal['l1', 'l2', 'elastic-net'] = None,
                 verbose: bool = False) -> None:
        self.alpha = alpha
        self.beta = beta
        self.learning_rate = learning_rate
        self.max_iter = max_iter
        self.reg_strength = reg_strength
        self.rho = rho
        self.regularization = regularization
        self.verbose = verbose
        self._fitted = False
    
    def fit(self, X: Matrix, y: Matrix) -> 'GammaRegressor':
        X = np.column_stack((np.ones(X.shape[0]), X))
        m, n = X.shape
        self.alpha = np.ones(n) * self.alpha
        self.beta = np.ones(n) * self.beta
        
        for i in range(self.max_iter):
            gradient = (self.alpha - 1) / self.alpha - psi(self.alpha)
            gradient += np.log(self.beta) - (y / self.beta)
            
            self.alpha += self.learning_rate * np.mean(gradient * X, axis=0) / m
            self.beta += self.learning_rate / m * \
                np.mean((self.alpha / self.beta ** 2 - y / self.beta) * X, axis=0)
            
            if self.regularization:
                self.alpha -= self.reg_strength * self._regularization_term(self.alpha) / m
                self.beta -= self.reg_strength * self._regularization_term(self.beta) / m
            
            if self.verbose and i % 10 == 0:
                print(f'[GammaReg] iteration: {i}/{self.max_iter}', end='')
                print(f' - gradient-norm: {np.linalg.norm(gradient)}')
        
        self._fitted = True
        return self
    
    def _regularization_term(self, weights: Matrix) -> Matrix:
        if self.regularization == 'l1': return np.sign(weights)
        elif self.regularization == 'l2': return self.weights
        elif self.regularization == 'elastic-net':
            l1_term = np.sign(weights)
            l2_term = 2 * weights
            return (1 - self.rho) * l2_term + self.rho * l1_term
        elif self.regularization is None: return np.zeros_like(weights)
        else: raise UnsupportedParameterError(self.regularization)

    def predict(self, X: Matrix) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        X = np.column_stack((np.ones(X.shape[0]), X))
        return np.dot(X, self.alpha / self.beta)
    
    def score(self, X: Matrix, y: Matrix, 
              metric: Evaluator = MeanSquaredError) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)
    
    def set_param(self,
                  alpha: float = None,
                  beta: float = None,
                  learning_rate: float = None,
                  max_iter: int = None,
                  reg_strength: float = None,
                  rho: float = None,
                  regularization: Literal = None) -> None:
        if alpha is not None: self.alpha = float(alpha)
        if beta is not None: self.beta = float(beta)
        if learning_rate is not None: self.learning_rate = float(learning_rate)
        if max_iter is not None: self.max_iter = int(max_iter)
        if reg_strength is not None: self.reg_strength = float(reg_strength)
        if rho is not None: self.rho = float(rho)
        if regularization is not None: self.regularization = str(regularization)


class BetaRegressor(Estimator, Supervised):
    
    """
    Beta regression is a statistical model used in machine learning to 
    analyze and predict data that follows a bounded interval between 
    0 and 1, such as proportions, probabilities, or percentages. 
    It's particularly useful for modeling data with heteroscedasticity, 
    where the spread of the data varies across different input values.
    
    Parameters
    ----------
    `alpha`, `beta` : Shape parameters of beta function
    `learning_rate` : Step size of the gradient descent update
    `max_iter` : Number of iteration
    `reg_strength` : Regularization strength
    `rho` : Balacing parameter of `elastic-net`
    `regularization` : Regularization type (e.g. `l1`, `l2`, `elastic-net`)
    
    """

    def __init__(self,
                 alpha: float = 1.0,
                 beta: float = 1.0,
                 learning_rate: float = 0.01,
                 max_iter: int = 100,
                 reg_strength: float = 0.01,
                 rho: float = 0.5,
                 regularization: Literal['l1', 'l2', 'elastic-net'] = None,
                 verbose: bool = False) -> None:
        self.alpha = alpha
        self.beta = beta
        self.learning_rate = learning_rate
        self.max_iter = max_iter
        self.reg_strength = reg_strength
        self.rho = rho
        self.regularization = regularization
        self.verbose = verbose
        self._fitted = False
    
    def fit(self, X: Matrix, y: Matrix) -> 'BetaRegressor':
        X = np.column_stack((np.ones(X.shape[0]), X))
        m, n = X.shape
        self.alpha = np.ones(n) * self.alpha
        self.beta = np.ones(n) * self.beta
        
        for i in range(self.max_iter):
            gradient_alpha = psi(self.alpha) - psi(self.alpha + self.beta)
            gradient_alpha += np.log(self.beta) - np.log(1 - y)
            gradient_beta = psi(self.beta) - psi(self.alpha + self.beta)
            gradient_beta += np.log(self.alpha) - np.log(y)
            
            self.alpha -= self.learning_rate * np.mean(gradient_alpha * X, axis=0) / m
            self.beta -= self.learning_rate * np.mean(gradient_beta * X, axis=0) / m
            
            if self.regularization:
                self.alpha -= self.reg_strength * self._regularization_term(self.alpha) / m
                self.beta -= self.reg_strength * self._regularization_term(self.beta) / m
            
            if self.verbose and i % 10 == 0:
                gradient_norm = np.linalg.norm(gradient_alpha + gradient_beta)
                print(f'[BetaReg] iteration: {i}/{self.max_iter}', end='')
                print(f' - gradient-norm: {gradient_norm}')
        
        self._fitted = True
        return self
    
    def _regularization_term(self, weights: Matrix) -> Matrix:
        if self.regularization == 'l1': return np.sign(weights)
        elif self.regularization == 'l2': return self.weights
        elif self.regularization == 'elastic-net':
            l1_term = np.sign(weights)
            l2_term = 2 * weights
            return (1 - self.rho) * l2_term + self.rho * l1_term
        elif self.regularization is None: return np.zeros_like(weights)
        else: raise UnsupportedParameterError(self.regularization)
    
    def predict(self, X: Matrix) -> float:
        if not self._fitted: raise NotFittedError(self)
        X = np.column_stack((np.ones(X.shape[0]), X))
        return np.dot(X, self.alpha / (self.alpha + self.beta))
    
    def score(self, X: Matrix, y: Matrix, 
              metric: Evaluator = MeanSquaredError) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)
    
    def set_params(self, 
                   alpha: float = None, 
                   beta: float = None, 
                   learning_rate: float = None, 
                   max_iter: int = None, 
                   reg_strength: float = None, 
                   rho: float = None, 
                   regularization: Literal = None) -> None:
        if alpha is not None: self.alpha = float(alpha)
        if beta is not None: self.beta = float(beta)
        if learning_rate is not None: self.learning_rate = float(learning_rate)
        if max_iter is not None: self.max_iter = int(max_iter)
        if reg_strength is not None: self.reg_strength = float(reg_strength)
        if rho is not None: self.rho = float(rho)
        if regularization is not None: self.regularization = str(regularization)


class InverseGaussianRegressor(Estimator, Supervised):
    
    """
    Inverse Gaussian regression is a statistical technique applied in 
    machine learning to model relationships between predictor variables 
    and continuous response variables when the response data exhibits 
    a skewed or non-normally distributed pattern.
    
    Parameters
    ----------
    `learning_rate` : Step size of the gradient descent update
    `max_iter` : Number of iteration
    `phi` : Shape parameter of inverse Gaussian density
    `rho` : Balacing parameter of `elastic-net`
    `alpha` : Regularization strength
    `regularization` : Regularization type (e.g. `l1`, `l2`, `elastic-net`)
    
    """
    
    def __init__(self,
                 learning_rate: float = 0.01,
                 max_iter: int = 100,
                 phi: float = 1.0,
                 rho: float = 0.5,
                 alpha: float = 0.01,
                 regularization: Literal['l1', 'l2', 'elastic-net'] = None,
                 verbose: bool = False) -> None:
        self.learning_rate = learning_rate
        self.max_iter = max_iter
        self.phi = phi
        self.rho = rho
        self.alpha = alpha
        self.regularization = regularization
        self.verbose = verbose
        self._fitted = False
    
    def fit(self, X: Matrix, y: Matrix) -> 'InverseGaussianRegressor':
        X = np.column_stack((np.ones(X.shape[0]), X))
        m, n = X.shape
        self.weights = np.ones(n)
        
        for i in range(self.max_iter):
            error = y - np.dot(X, self.weights)
            
            gradient = np.dot(X.T, (error * self.phi) / (2 * self.weights * y ** 2))
            gradient += 0.5 * error ** 2 / (2 * self.weights * y)
            gradient -= 0.5 * (self.phi / self.weights)
            gradient *= self.learning_rate / m
            
            if self.regularization:
                regularization_term = self._regularization_term()
                gradient += self.alpha * regularization_term / m
                
            if self.verbose and i % 10 == 0:
                print(f'[InvGaussianReg] iteration: {i}/{self.max_iter}', end='')
                print(f' - gradient-norm: {np.linalg.norm(gradient)}')
            
            self.weights -= gradient
        self._fitted = True
        return self
    
    def _regularization_term(self):
        if self.regularization == 'l1': return np.sign(self.weights)
        elif self.regularization == 'l2': return self.weights
        elif self.regularization == 'elastic-net':
            l1_term = np.sign(self.weights)
            l2_term = 2 * self.weights
            return (1 - self.rho) * l2_term + self.rho * l1_term
        elif self.regularization is None: return np.zeros_like(self.weights)
        else: raise UnsupportedParameterError(self.regularization)
    
    def predict(self, X: Matrix) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        X = np.column_stack((np.ones(X.shape[0]), X))
        return 1 / (self.phi * X) * np.dot(X, self.weights)
    
    def score(self, X: Matrix, y: Matrix, 
              metric: Evaluator = MeanSquaredError) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)

    def set_params(self, 
                   learning_rate: float = None, 
                   max_iter: int = None, 
                   phi: float = None,
                   rho: float = None,
                   alpha: float = None, 
                   regularization: Literal = None) -> None:
        if learning_rate is not None: self.learning_rate = float(learning_rate)
        if max_iter is not None: self.max_iter = int(max_iter)
        if phi is not None: self.phi = float(phi)
        if rho is not None: self.rho = float(rho)
        if alpha is not None: self.alpha = float(alpha)
        if regularization is not None: self.regularization = str(regularization)
        
        


from typing import Any, Callable, Tuple
import numpy as np

from luma.interface.super import Estimator, Transformer
from luma.interface.exception import UnsupportedParameterError


__all__ = (
    'Matrix', 
    'Vector', 
    'Scalar', 
    'TreeNode', 
    'NearestNeighbors', 
    'SilhouetteUtil', 
    'DBUtil',
    'KernelUtil',
    'Clone'
)


class Matrix(np.ndarray):

    """
    Internal class that extends `numpy.ndarray`.

    This class provides a way to create matrix objects that have 
    all the capabilities of numpy arrays with the potential for 
    additional functionalities and readability.s
    
    Example
    -------
    >>> m = Matrix([1, 2, 3])
    >>> isinstance(m, numpy.ndarray) # True
    
    """
    
    def __new__(cls, array_like: Any) -> 'Matrix':
        if isinstance(array_like, list): obj = np.array(array_like)
        else: obj = array_like    
        return obj

    def __array_finalize__(self, obj: np.ndarray | None) -> None:
        if obj is None: return


class Vector(Matrix):
    
    """
    Internal class for vector that extends `Matrix`.
    
    This class represents a single row/column vector with its
    type of `numpy.ndarray` or `Matrix`.
    
    """
    
    def __new__(cls, array_like: Any) -> 'Vector':
        if isinstance(array_like, list): obj = Matrix(array_like)
        else: obj = array_like    
        return obj


class Scalar:
    
    """
    A placeholder class for scalar type.
    
    This class encompasses `int` and `float`.
    """
    
    def __new__(cls, value: int | float) -> 'Scalar':
        return float(value)


class TreeNode:
    
    """
    Internal class for node used in tree-based models.
    
    Parameters
    ----------
    `feature` : Feature of node
    `threshold` : Threshold for split point
    `left` : Left-child node
    `right` : Right-child node
    `value` : Most popular label of leaf node
    
    """
    
    def __init__(self,
                 feature: int = None,
                 threshold: float = None,
                 left: 'TreeNode' = None,
                 right: 'TreeNode' = None,
                 value: int | float = None) -> None:
        self.feature = feature
        self.threshold = threshold
        self.left = left
        self.right = right
        self.value = value

    @property
    def isLeaf(self) -> bool:
        return self.value is not None


class NearestNeighbors:
    
    """
    Internal class for computing nearest neighbors of given data.
    
    Parameters
    ----------
    `data` : Data to be handled
    `n_neighbors` : Number of nearest neighbors
    
    """
    
    def __init__(self,
                 data: Matrix,
                 n_neighbors: int) -> None:
        self.data = data
        self.n_neighbors = n_neighbors
        self._size = data.shape[0]
    
    @property
    def index_matrix(self) -> Matrix:
        data = self.data
        dist = np.linalg.norm(data[:, np.newaxis, :] - data, axis=2)
        sorted_indices = np.argsort(dist, axis=1)
        return sorted_indices[:, 1:self.n_neighbors + 1]
    
    @property
    def adjacent_matrix(self) -> Matrix:
        indices = self.index_matrix
        adj_mat = np.zeros((self._size, self._size))
        for i in range(self._size):
            adj_mat[i, indices[i]] = 1
        
        return adj_mat.astype(int)


class SilhouetteUtil:
    
    """
    Internal class for computing various distances used in 
    Silhouette Coefficient calculation.
    
    Parameters
    ----------
    `idx` : Index of a single data point
    `cluster` : Current cluster number
    `labels` : Labels assigned by clustering estimator
    `distances` : Square-form distance matrix of the data
    
    """
    
    def __init__(self, 
                 idx: int,
                 cluster: int,
                 labels: Matrix,
                 distances: Matrix) -> None:
        self.idx = idx
        self.cluster = cluster
        self.labels = labels
        self.distances = distances

    @property
    def avg_dist_others(self) -> Matrix:
        others = set(self.labels) - {self.cluster}
        sub_avg = [np.mean(self.distances[self.idx][self.labels == other]) 
                   for other in others]

        return np.mean(sub_avg)

    @property
    def avg_dist_within(self) -> Matrix | int:
        within_cluster = self.distances[self.idx][self.labels == self.cluster]
        if len(within_cluster) <= 1: return 0
        return np.mean([dist for dist in within_cluster if dist != 0])


class DBUtil:
    
    """
    Internal class for supporting Davies-Bouldin Index (DBI) computation.
    
    Parameters
    ----------
    `data` : Original data
    `labels` : Labels assigned by clustering estimator
    
    """
    
    def __init__(self,
                 data: Matrix, 
                 labels: Matrix) -> None:
        self.data = data
        self.labels = labels
    
    @property
    def cluster_centroids(self) -> Matrix:
        unique_labels = np.unique(self.labels)
        centroids = np.array([self.data[self.labels == label].mean(axis=0) 
                              for label in unique_labels])
        return centroids

    @property
    def within_cluster_scatter(self) -> Matrix:
        centroids = self.cluster_centroids
        scatter = np.zeros(len(centroids))
        
        for i, centroid in enumerate(centroids):
            cluster_points = self.data[self.labels == i]
            diff_sq =  (cluster_points - centroid) ** 2
            scatter[i] = np.mean(np.sqrt(np.sum(diff_sq, axis=1)))
            
        return scatter

    @property
    def separation(self) -> Matrix:
        centroids = self.cluster_centroids
        n_clusters = len(centroids)
        separation = np.zeros((n_clusters, n_clusters))
        
        for i in range(n_clusters):
            for j in range(n_clusters):
                if i == j: continue
                diff_sq = (centroids[i] - centroids[j]) ** 2
                separation[i, j] = np.sqrt(np.sum(diff_sq))
                    
        return separation


class KernelUtil:
    
    """
    Internal class for kernel methods(tricks).
    
    This class facilitates transferring kernel type strings
    into actual specific kernel function.
    
    Example
    -------
    >>> util = KernelUtil(kernel='rbf')
    >>> util.kernel_func
    KernelUtil.rbf_kernel: Callable[[...], Matrix]
    
    """
    
    def __init__(self, kernel: str) -> None:
        self.kernel = kernel
    
    def linear_kernel(self, X: Matrix) -> Matrix:
        return np.dot(X, X.T)
    
    def polynomial_kernel(self, X: Matrix, 
                          gamma: float = 1.0,
                          coef: float = 0.0,
                          deg: int = 2) -> Matrix:
        return (gamma * np.dot(X, X.T) + coef) ** deg
    
    def rbf_kernel(self, X: Matrix, gamma: float = 1.0) -> Matrix:
        _left = np.sum(X ** 2, axis=1).reshape(-1, 1)
        _right = np.sum(X ** 2, axis=1) - 2 * np.dot(X, X.T)
        
        return np.exp(-gamma * (_left + _right))
    
    def sigmoid_kernel(self, X: Matrix,
                       gamma: float = 1.0,
                       coef: float = 0.0) -> Matrix:
        return np.tanh(gamma * np.dot(X, X.T) + coef)
    
    def laplacian_kernel(self, X: Matrix, gamma: float = 1.0) -> Matrix:
        manhattan_dists = np.sum(np.abs(X[:, np.newaxis] - X), axis=2)
        return np.exp(-gamma * manhattan_dists)
    
    @property
    def kernel_func(self) -> Callable[[Matrix, Ellipsis], Matrix]:
        if self.kernel in ('linear', 'lin'):
            return self.linear_kernel
        elif self.kernel in ('poly', 'polynomial'):
            return self.polynomial_kernel
        elif self.kernel in ('rbf', 'gaussian', 'Gaussian'):
            return self.rbf_kernel
        elif self.kernel in ('sigmoid', 'tanh'):
            return self.sigmoid_kernel
        elif self.kernel in ('laplacian', 'lap'):
            return self.laplacian_kernel
        else:
            raise UnsupportedParameterError(self.kernel)


class Clone:
    
    """
    A utility class for cloning LUMA models.

    This class creates a copy of a given LUMA model, 
    which can be either an Estimator or a Transformer. 
    The clone includes all parameters of the original model. 
    Optionally, the trained state of the model can also be copied 
    if applicable.

    Parameters
    ----------
    `model` : The model to be cloned
    `pass_fitted` : Whether to copy the fitted state of the original model
    
    Examples
    --------
    >>> original_model = AnyModel(...)
    >>> cloned_model = Clone(model=original_model, pass_fitted=True).get

    """
    
    def __init__(self,
                 model: Estimator | Transformer = None,
                 pass_fitted: bool = False) -> None:
        self.model = model
        self.pass_fitted = pass_fitted
    
    @property
    def get(self) -> Estimator | Transformer:
        model_cls = type(self.model)
        new_model = model_cls()
        
        for param, val in self.model.__dict__.items():
            try: new_model.set_params(**{param: val})
            except: continue
        
        if hasattr(self.model, '_fitted') and self.pass_fitted:
            new_model._fitted = self.model._fitted
        
        return new_model




from typing import *
import numpy as np

from luma.interface.util import Matrix
from luma.interface.exception import NotFittedError
from luma.interface.super import Estimator, Evaluator
from luma.metric.regression import MeanSquaredError


__all__ = (
    'PolynomialRegressor'
)


class PolynomialRegressor(Estimator):
    
    """
    Polynomial regression is a type of regression analysis used
    in statistics and machine learning to model the relationship between 
    a dependent variable and one or more independent variables.
    """
    
    def __init__(self, degree: int = 1):
        self.degree = degree
        self._fitted = False

    def _generate_polynomial_features(self, X):
        X_poly = X.copy()
        for d in range(2, self.degree + 1):
            X_poly = np.hstack((X_poly, X ** d))
        return X_poly

    def fit(self, X: Matrix, y: Matrix) -> 'PolynomialRegressor':
        X_poly = self._generate_polynomial_features(X)
        self.coefficients = np.linalg.lstsq(X_poly, y, rcond=None)[0]
        self._fitted = True
        return self

    def predict(self, X: Matrix) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        X_poly = self._generate_polynomial_features(X)
        return np.dot(X_poly, self.coefficients)

    def score(self, X: Matrix, y: Matrix, 
              metric: Evaluator = MeanSquaredError) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)

    def set_params(self, degree: int = None) -> None:
        self.degree = int(degree)




import __local__
from luma.model_selection.robust import RANSAC, MLESAC
from luma.regressor.poly import PolynomialRegressor
from luma.visual.result import ResidualPlot

import matplotlib.pyplot as plt
import numpy as np


np.random.seed(10)
X = np.linspace(-3, 3, 1000)
y = X ** 3

indices = np.arange(650, 800)
y[indices] = -10 * X[indices]
y += np.random.normal(0, 1.5, 1000)
X = X.reshape(-1, 1)

model = MLESAC(estimator=PolynomialRegressor(degree=3),
               min_points=2,
               max_iter=1000,
               min_inliers=0.5,
               threshold=None,
               random_state=10)

model.fit(X, y)

coef = model.best_estimator.coefficients
func = f"({coef[2]:.3f}){r'$x^3$'}+({coef[1]:.3f}){r'$x^2$'}+({coef[0]:.3f}){r'$x$'}"

res = ResidualPlot(estimator=model,
                   X=X,
                   y=y)

res.plot(show=True)



from typing import List, Literal
import numpy as np

from luma.interface.super import Estimator, Evaluator, Supervised
from luma.interface.util import Matrix, Vector, Clone
from luma.interface.exception import NotFittedError, UnsupportedParameterError
from luma.classifier.tree import DecisionTreeClassifier
from luma.regressor.tree import DecisionTreeRegressor
from luma.preprocessing.encoder import LabelBinarizer
from luma.metric.classification import Accuracy
from luma.metric.regression import MeanSquaredError


__all__ = (
    'AdaBoostClassifier',
    'AdaBoostRegressor'
)


class AdaBoostClassifier(Estimator, Estimator.Meta, Supervised):
    
    """
    The AdaBoost classifier is an ensemble method that combines multiple 
    weak learners, typically decision trees, to form a strong classifier. 
    During training, it iteratively adjusts the weights of misclassified 
    instances, making them more likely to be correctly predicted in subsequent 
    rounds. Each weak learner's contribution to the final prediction is 
    weighted based on its accuracy. AdaBoost is adaptive in the sense that 
    subsequent weak learners are tweaked in favor of instances that previous 
    learners misclassified. The final model makes predictions based on the 
    weighted majority vote of its weak learners.
    
    Parameters
    ----------
    `base_estimator` : Base estimator for training multiple models
    (Default `DecisionTreeClassifier`)
    `n_estimators` : Number of base estimators to fit
    `learning_rate` : Step size of class weights(`alpha`) update
    
    Examples
    --------
    >>> ada_clf = AdaBoostClassifier(learning_rate=1.0, ...)
    >>> ada_clf.fit(X, y)
    
    >>> pred = ada_clf.predict(X)
    >>> est = ada_clf[i] # Get i-th estimator from `ada_clf`
    
    """
    
    def __init__(self, 
                 base_estimator: Estimator = DecisionTreeClassifier(),
                 n_estimators: int = 100, 
                 learning_rate: float = 1.0,
                 verbose: bool = False) -> None:
        self.base_estimator = base_estimator
        self.n_estimators = n_estimators
        self.learning_rate = learning_rate
        self.verbose = verbose
        self.estimators_: List[Estimator] = []
        self.weights_: List[float] = []
        self.classes_ = None
        self._fitted = False

    def fit(self, X: Matrix, y: Matrix) -> 'AdaBoostClassifier':
        self.classes_ = np.unique(y)
        m, _ = X.shape
        n_classes = len(self.classes_)
        sample_weights = np.full(m, 1 / m)
        
        lb = LabelBinarizer().fit(self.classes_)
        y_bin = lb.transform(y)

        for i in range(self.n_estimators):
            clf = Clone(self.base_estimator).get
            if hasattr(clf, 'max_depth'): clf.max_depth = 1
            if hasattr(clf, 'sample_weights'): clf.sample_weights = sample_weights
            
            clf.fit(X, y)
            pred = clf.predict(X)
            pred_bin = lb.transform(pred)

            error_vect = np.sum(y_bin != pred_bin, axis=1) / (n_classes - 1)
            error = np.sum(sample_weights * error_vect) / np.sum(sample_weights)
            error = max(error, 1e-10)
            if error >= (n_classes - 1) / n_classes: continue

            alpha = self.learning_rate * np.log((1 - error) / error)
            sample_weights *= np.exp(alpha * error_vect)
            sample_weights /= np.sum(sample_weights)

            self.estimators_.append(clf)
            self.weights_.append(alpha)
            
            if self.verbose:
                print(f'[AdaBoost] Finished fitting {type(clf).__name__}',
                      f'{i + 1}/{self.n_estimators} with',
                      f'error: {error}, alpha: {alpha}')

        self._fitted = True
        return self

    def predict(self, X: Matrix) -> Vector:
        if not self._fitted: raise NotFittedError(self)
        m, _ = X.shape
        clf_preds = np.array([clf.predict(X) for clf in self.estimators_])
        weighted_preds = np.zeros((m, len(self.classes_)))
        
        for i, clf_pred in enumerate(clf_preds):
            for j in range(m):
                class_index = np.where(self.classes_ == clf_pred[j])[0][0]
                weighted_preds[j, class_index] += self.weights_[i]
            
            if self.verbose:
                print(f'[AdaBoost] Finished prediction of',
                      f'{type(self.base_estimator).__name__}',
                      f'{i + 1}/{self.n_estimators}')
        
        return self.classes_[np.argmax(weighted_preds, axis=1)]

    def score(self, X: Matrix, y: Vector, metric: Evaluator = Accuracy) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)
    
    def set_params(self, 
                   base_estimator: Estimator = None,
                   n_estimators: int = None,
                   learning_rate: float = None) -> None:
        if base_estimator is not None: self.base_estimator = base_estimator
        if n_estimators is not None: self.n_estimators = n_estimators
        if learning_rate is not None: self.learning_rate = learning_rate

    def __getitem__(self, index: int) -> Estimator:
        return self.estimators_[index]


class AdaBoostRegressor(Estimator, Estimator.Meta, Supervised):
    
    """
    The AdaBoost regressor is an ensemble method that combines multiple 
    weak learners, typically decision trees, to form a strong regressor. 
    During training, it focuses on reducing the errors of the ensemble by 
    adjusting the sample weights and fitting each weak learner to the 
    residual errors of the previous learners. The final output is a 
    combination of the weak learners' outputs, weighted by their individual 
    contributions to the ensemble's performance, which aims to produce 
    a more accurate regression model.
    
    Parameters
    ----------
    `base_estimator` : Base estimator for training multiple models
    (Default `DecisionTreeRegressor`)
    `n_estimators` : Number of base estimators to fit
    `learning_rate` : Step size of class weights(`alpha`) update
    `loss` : Type of loss function (e.g. `linear`, `square`, `exp`)

    Examples
    --------
    >>> ada_reg = AdaBoostRegressor(learning_rate=0.1, ...)
    >>> ada_reg.fit(X_train, y_train)

    >>> y_pred = ada_reg.predict(X_test)
    >>> est = ada_reg[i] # Get the i-th estimator from the `ada_reg`

    """
    
    def __init__(self, 
                 base_estimator: Estimator = DecisionTreeRegressor(),
                 n_estimators: int = 100,
                 learning_rate: float = 1.0,
                 loss: Literal['linear', 'sqaure', 'exp'] = 'linear',
                 verbose: bool = False) -> None:
        self.base_estimator = base_estimator
        self.n_estimators = n_estimators
        self.learning_rate = learning_rate
        self.loss = loss
        self.verbose = verbose
        self.estimators_: List[Estimator] = []
        self.weights_: List[float] = []
        self._fitted = False
    
    def fit(self, X: Matrix, y: Vector) -> 'AdaBoostRegressor':
        m, _ = X.shape
        sample_weights = np.full(m, 1 / m)
        y_pred = np.zeros(m)
        
        for i in range(self.n_estimators):
            residual = y - y_pred
            clf = Clone(self.base_estimator).get
            if hasattr(clf, 'max_depth'): clf.max_depth = 1
            if hasattr(clf, 'sample_weights'): clf.sample_weights = sample_weights
            
            clf.fit(X, residual)
            pred = clf.predict(X)
            loss = self._compute_loss(residual, pred)
            
            error = max(np.sum(sample_weights * loss), 1e-10)
            alpha = self.learning_rate * np.log((1 - error) / error)
            
            sample_weights *= np.exp(alpha * loss)
            sample_weights /= np.sum(sample_weights)
            
            y_pred += alpha * pred
            self.estimators_.append(clf)
            self.weights_.append(alpha)
            
            if self.verbose:
                print(f'[AdaBoost] Finished fitting {type(clf).__name__}',
                      f'{i + 1}/{self.n_estimators} with',
                      f'error: {error}, alpha: {alpha}')
        
        self._fitted = True
        return self
    
    def _compute_loss(self, y: Vector, y_pred: Vector) -> Vector:
        if self.loss == 'linear': return np.abs(y - y_pred)
        elif self.loss == 'square': return (y - y_pred) ** 2
        elif self.loss == 'exp': return np.exp(-y * y_pred)
        else: raise UnsupportedParameterError(self.loss)
    
    def predict(self, X: Matrix) -> Vector:
        if not self._fitted: raise NotFittedError(self)
        predictions = np.zeros(X.shape[0])
        for weight, estimator in zip(self.weights_, self.estimators_):
            predictions += weight * estimator.predict(X)
        
        return predictions
    
    def score(self, X: Matrix, y: Vector, 
              metric: Evaluator = MeanSquaredError) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)

    def set_params(self, 
                   base_estimator: Estimator = None,
                   n_estimators: int = None,
                   learning_rate: float = None,
                   loss: str = None) -> None:
        if base_estimator is not None: self.base_estimator = base_estimator
        if n_estimators is not None: self.n_estimators = int(n_estimators)
        if learning_rate is not None: self.learning_rate = float(learning_rate)
        if loss is not None: self.loss = loss
    
    def __getitem__(self, index: int) -> Estimator:
        return self.estimators_[index]
    



from typing import Literal
from scipy.spatial.distance import pdist, squareform
from scipy.linalg import eigh
import numpy as np

from luma.clustering.kmeans import KMeansClusteringPlus
from luma.clustering.hierarchy import AgglomerativeClustering, DivisiveClustering
from luma.interface.util import Matrix
from luma.interface.super import Estimator, Evaluator, Unsupervised
from luma.interface.exception import NotFittedError, UnsupportedParameterError
from luma.metric.clustering import SilhouetteCoefficient


__all__ = (
    'SpectralClustering', 
    'NormalizedSpectralClustering',
    'HierarchicalSpectralClustering',
    'AdaptiveSpectralClustering'
)


class SpectralClustering(Estimator, Unsupervised):
    
    """
    Spectral clustering transforms data into a new space using the 
    eigenvectors of its graph Laplacian, revealing cluster structure. 
    It constructs a similarity graph from the data, capturing complex 
    relationships between points. The algorithm then applies 
    dimensionality reduction followed by a traditional clustering 
    method, like K-means, in this transformed space. This approach 
    is effective for identifying clusters with irregular shapes 
    and non-linearly separable data.
    
    Parameters
    ----------
    `n_clusters` : Number of clusters to estimate
    `gamma` : Scaling factor for Gaussian kernel
    
    Examples
    --------
    >>> sp = SpectralClustering()
    >>> sp.fit(X, y)
    >>> lables = sp.labels # Get assigned labels
    
    """
    
    def __init__(self, 
                 n_clusters: int, 
                 gamma: float = 1.0):
        self.n_clusters = n_clusters
        self.gamma = gamma
        self._X = None
        self._fitted = False
    
    def fit(self, X: Matrix) -> 'SpectralClustering':
        self._X = X
        W = self._similarity_matrix(X)
        L = self._laplacian(W)
        
        _, eigenvecs = eigh(L)
        V = eigenvecs[:, :self.n_clusters]
        kmeans = KMeansClusteringPlus(n_clusters=self.n_clusters)
        self.kmeans = kmeans.fit(V)
        
        self._fitted = True
        return self

    def _similarity_matrix(self, X: Matrix) -> Matrix:
        sq_dists = pdist(X, 'sqeuclidean')
        mat_sq_dists = squareform(sq_dists)
        return np.exp(-self.gamma * mat_sq_dists)

    def _laplacian(self, W: Matrix) -> Matrix:
        D = np.diag(np.sum(W, axis=1))
        return D - W

    @property
    def labels(self) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        return self.kmeans.labels
    
    def predict(self) -> None:
        raise Warning(f"{type(self).__name__} does not support prediction!")
    
    def score(self, metric: Evaluator = SilhouetteCoefficient) -> float:
        return metric.score(self._X, self.labels)
    
    def set_params(self, 
                   n_clusters: int = None,
                   gamma: float = None) -> None:
        if n_clusters is not None: self.n_clusters = int(n_clusters)
        if gamma is not None: self.gamma = float(gamma)


class NormalizedSpectralClustering(Estimator, Unsupervised):
    
    """
    Normalized spectral clustering is a technique that begins by constructing 
    a similarity matrix from the data, often using a Gaussian kernel. The graph 
    Laplacian matrix, derived from this similarity matrix, is then normalized 
    to enhance the distinctiveness of clusters. Eigenvalue decomposition is 
    applied to this normalized Laplacian, and the resulting eigenvectors are 
    used to transform the data into a new space. Finally, a clustering algorithm, 
    such as k-means, is applied in this transformed space, effectively revealing 
    clusters that are not easily distinguishable in the original space.
    
    Parameters
    ----------
    `n_clusters` : Number of clusters to estimate
    `gamma` : Scaling factor for Gaussian kernel
    `strategy` : Normalization strategy (e.g. `symmetric`, `random-walk`)
    
    """
    
    def __init__(self, 
                 n_clusters: int,
                 gamma: float = 1.0,
                 strategy: Literal['symmetric', 
                                   'random-walk'] = 'symmetric') -> None:
        self.n_clusters = n_clusters
        self.gamma = gamma
        self.strategy = strategy
        self._X = None
        self._fitted = False
    
    def fit(self, X: Matrix) -> 'NormalizedSpectralClustering':
        self._X = X
        W = self._similarity_matrix(X)
        L_norm = self._normalize_laplacian(W)
        
        _, eigvecs = eigh(L_norm)
        V = eigvecs[:, :self.n_clusters]
        kmeans = KMeansClusteringPlus(n_clusters=self.n_clusters)
        self.kmeans = kmeans.fit(V)
        
        self._fitted = True
        return self
    
    def _similarity_matrix(self, X: Matrix) -> Matrix:
        sq_dists = pdist(X, 'sqeuclidean')
        mat_sq_dists = squareform(sq_dists)
        return np.exp(-self.gamma * mat_sq_dists)
    
    def _normalize_laplacian(self, W: Matrix) -> Matrix:
        D = np.diag(np.sum(W, axis=1))
        D_inv = np.diag(1 / np.diag(D))
        D_inv_sqrt = np.diag(1 / np.sqrt(np.diag(D)))
        
        if self.strategy == 'symmetric':
            L_norm = np.eye(*D.shape) - D_inv_sqrt.dot(W).dot(D_inv_sqrt)
        elif self.strategy == 'random-walk':
            L_norm = np.eye(*D.shape) - D_inv.dot(W)
        else:
            raise UnsupportedParameterError(self.strategy)

        return L_norm
    
    @property
    def labels(self) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        return self.kmeans.labels

    def predict(self) -> None:
        raise Warning(f"{type(self).__name__} does not support prediction!")
    
    def score(self, metric: Evaluator = SilhouetteCoefficient) -> float:
        return metric.score(self._X, self.labels)
    
    def set_params(self, 
                   n_clusters: int = None,
                   gamma: float = None,
                   strategy: Literal = None) -> None:
        if n_clusters is not None: self.n_clusters = int(n_clusters)
        if gamma is not None: self.gamma = float(gamma)
        if strategy is not None: self.strategy = str(strategy)


class HierarchicalSpectralClustering(Estimator, Unsupervised):
    
    """
    Hierarchical spectral clustering starts by applying spectral clustering 
    to transform data into a space where clusters are more linearly separable. 
    This involves constructing a similarity matrix, computing the Laplacian, 
    and performing eigenvalue decomposition. The transformed data is then 
    clustered using hierarchical methods, either agglomerative (building 
    clusters by merging smaller ones) or divisive (splitting larger clusters 
    into smaller ones). This approach reveals multi-level structures in the 
    data, uncovering both broad and fine-grained clustering patterns.
    
    Parameters
    ----------
    `n_clusters` : Number of clusters to estimate
    `method` : Method for hierarchical clustering
    (e.g. `agglomerative`, `divisive`)
    `linkage` : Linkage method for `agglomerative` clustering
    (e.g. `single`, `complete`, `average`, `ward`)
    `gamma` : Scaling factor for Gaussian kernel
    
    """
    
    def __init__(self, 
                 n_clusters: int, 
                 method: Literal['agglomerative', 'divisive'],
                 linkage: Literal['single', 'complete', 
                                  'average', 'ward'] = 'single',
                 gamma: float = 1.0,):
        self.n_clusters = n_clusters
        self.method = method
        self.linkage = linkage
        self.gamma = gamma
        self._hierarchy_model = None
        self._X = None
        self._fitted = False
    
    def fit(self, X: Matrix) -> 'HierarchicalSpectralClustering':
        self._X = X
        W = self._similarity_matrix(X)
        L = self._laplacian(W)
        
        _, eigenvecs = eigh(L)
        V = eigenvecs[:, :self.n_clusters]
        
        if self.method == 'agglomerative':
            self._hierarchy_model = AgglomerativeClustering()
        elif self.method == 'divisive':
            self._hierarchy_model = DivisiveClustering()
        else:
            raise UnsupportedParameterError(self.method)
        
        self._hierarchy_model.n_clusters = self.n_clusters
        if hasattr(self._hierarchy_model, 'linkage'):
            self._hierarchy_model.linkage = self.linkage
        
        self._hierarchy_model.fit(V)
        
        self._fitted = True
        return self

    def _similarity_matrix(self, X: Matrix) -> Matrix:
        sq_dists = pdist(X, 'sqeuclidean')
        mat_sq_dists = squareform(sq_dists)
        return np.exp(-self.gamma * mat_sq_dists)

    def _laplacian(self, W: Matrix) -> Matrix:
        D = np.diag(np.sum(W, axis=1))
        return D - W

    @property
    def labels(self) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        return self._hierarchy_model.labels
    
    def predict(self) -> None:
        raise Warning(f"{type(self).__name__} does not support prediction!")
    
    def score(self, metric: Evaluator = SilhouetteCoefficient) -> float:
        return metric.score(self._X, self.labels)
    
    def set_params(self, 
                   n_clusters: int = None,
                   method: Literal = None,
                   linkage: Literal = None,
                   gamma: float = None) -> None:
        if n_clusters is not None: self.n_clusters = int(n_clusters)
        if method is not None: self.method = str(method)
        if linkage is not None: self.linkage = str(linkage)
        if gamma is not None: self.gamma = float(gamma)


class AdaptiveSpectralClustering(Estimator, Unsupervised):
    
    """
    Adaptive Spectral Clustering automatically determines the number of clusters 
    and scale parameters directly from the data, reducing the need for manual 
    tuning. It constructs a similarity matrix, typically using a Gaussian kernel 
    with an adaptively chosen scale. Eigenvalue decomposition on the Laplacian of 
    this matrix reveals the data's structure, and the eigen-gap heuristic is 
    employed to select the optimal number of clusters.
    
    Parameters
    ----------
    `gamma` : Scaling factor for Gaussian kernel
    `max_clusters` : Upper-bound of the number of clusters to estimate
    
    """
    
    def __init__(self, 
                 gamma: float = 1.0,
                 max_clusters: int = 10) -> None:
        self.gamma = gamma
        self.max_clusters = max_clusters
        self._X = None
        self._fitted = False
    
    def fit(self, X: Matrix) -> 'AdaptiveSpectralClustering':
        self._X = X
        W = self._similarity_matrix(X)
        L = self._laplacian(W)
        
        eigvals, eigvecs = eigh(L)
        self.n_clusters = self._optimal_clusters(eigvals)
        
        V = eigvecs[:, :self.n_clusters]
        kmeans = KMeansClusteringPlus(n_clusters=self.n_clusters)
        self.kmeans = kmeans.fit(V)
        
        self._fitted = True
        return self

    def _similarity_matrix(self, X: Matrix) -> Matrix:
        sq_dists = pdist(X, metric='sqeuclidean')
        mat_sq_dists = squareform(sq_dists)
        return np.exp(-self.gamma * mat_sq_dists)
    
    def _laplacian(self, W: Matrix) -> Matrix:
        D = np.diag(np.sum(W, axis=0))
        return D - W
    
    def _optimal_clusters(self, eigvals: Matrix) -> Matrix:
        eigval_diff = np.diff(eigvals)
        return np.argmax(eigval_diff[:self.max_clusters]) + 1
    
    @property
    def labels(self) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        return self.kmeans.labels
    
    def predict(self) -> None:
        raise Warning(f"{type(self).__name__} does not support prediction!")

    def score(self, metric: Evaluator = SilhouetteCoefficient) -> float:
        return metric.score(self._X, self.labels)
    
    def set_params(self, 
                   gamma: float = None,
                   max_clusters: int = None) -> None:
        if gamma is not None: self.gamma = float(gamma)
        if max_clusters is not None: self.max_clusters = int(max_clusters)




import numpy as np

from luma.interface.super import Estimator, Evaluator, Supervised
from luma.interface.util import Matrix, Vector, Clone
from luma.interface.exception import NotFittedError
from luma.classifier.tree import DecisionTreeClassifier
from luma.regressor.tree import DecisionTreeRegressor
from luma.metric.classification import Accuracy
from luma.metric.regression import MeanSquaredError


__all__ = (
    'BaggingClassifier',
    'BaggingRegressor'
)


class BaggingClassifier(Estimator, Estimator.Meta, Supervised):
    
    """
    A Bagging Classifier is an ensemble learning technique that trains 
    multiple base models, typically decision trees, on random subsets 
    of the original dataset. Each model is trained on a bootstrap sample 
    (random sample with replacement) of the data, potentially using a 
    random subset of features. The final prediction is made by aggregating 
    the predictions of all models, typically through voting for 
    classification or averaging for regression. This approach reduces 
    variance, improves model stability, and can prevent overfitting.
    
    Parameters
    ----------
    `base_estimator` : Base estimator for training multiple models
    (Default `DecisionTreeClassifier`)
    `n_estimators` : Number of base estimators to fit
    `max_samples` : Maximum number of data to sample (`0~1` proportion)
    `max_features` : Maximum number of features to sample (`0~1` proporton)
    `bootstrap` : Whether to bootstrap data samples
    `bootstrap_feature`: Whether to bootstrap features
    `random_state` : Seed for random sampling
    
    Examples
    --------
    >>> bag = BaggingClassifier(base_estimator=AnyEstimator(),
                                n_estimators=100,
                                max_samples=1.0,
                                max_features=1.0,
                                bootstrap=True,
                                bootstrap_feature=False)
    >>> bag.fit(X, y)
    >>> y_pred = bag.predict(X)
    >>> est = bag[i] # Get i-th base estimator from `bag`
    
    """
    
    def __init__(self, 
                 base_estimator: Estimator = DecisionTreeClassifier(),
                 n_estimators: int = 50,
                 max_samples: float | int = 1.0,
                 max_features: float | int = 1.0,
                 bootstrap: bool = True,
                 bootstrap_feature: bool = False,
                 random_state: int = None,
                 verbose: bool = False) -> None:
        self.base_estimator = base_estimator
        self.n_estimators = n_estimators
        self.max_samples = max_samples
        self.max_features = max_features
        self.bootstrap = bootstrap
        self.bootstrap_feature = bootstrap_feature
        self.random_state = random_state
        self.verbose = verbose
        self._fitted = False
    
    def fit(self, X: Matrix, y: Vector) -> 'BaggingClassifier':
        np.random.seed(self.random_state)
        self.estimators_ = []
        
        m, n = X.shape
        if self.max_samples <= 1.0: max_samples = int(m * self.max_samples)
        else: max_samples = self.max_samples
        
        if self.max_features <= 1.0: max_features = int(n * self.max_features)
        else: max_features = self.max_features
        
        for i in range(self.n_estimators):
            s_indices = np.random.choice(m, max_samples, self.bootstrap)
            f_indices = np.random.choice(n, max_features, self.bootstrap_feature)
            
            X_sample = X[s_indices][:, f_indices]
            y_sample = y[s_indices]
            
            estimator = Clone(self.base_estimator).get
            estimator.fit(X_sample, y_sample)
            self.estimators_.append((estimator, f_indices))
            
            if self.verbose:
                print(f'[Bagging] Finished fitting',
                      f'{type(self.base_estimator).__name__}',
                      f'{i + 1}/{self.n_estimators}')
        
        self._fitted = True
        return self
    
    def predict(self, X: Matrix) -> Vector:
        if not self._fitted: raise NotFittedError(self)
        
        predictions = []
        for i, (estimator, f_indices) in enumerate(self.estimators_):
            predictions.append(estimator.predict(X[:, f_indices]))
            if self.verbose:
                print(f'[Bagging] Finished prediction of',
                      f'{type(self.base_estimator).__name__}',
                      f'{i + 1}/{self.n_estimators}')
        
        majority_vote = np.apply_along_axis(
            lambda x: np.bincount(x, minlength=len(np.unique(x))).argmax(),
            axis=0,
            arr=predictions
        )
        return majority_vote
    
    def score(self, X: Matrix, y: Vector, metric: Evaluator = Accuracy) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)
    
    def set_params(self, 
                   base_estimator: Estimator = None,
                   n_estimators: int = None,
                   max_samples: float | int = None,
                   max_features: float | int = None,
                   bootstrap: bool = False,
                   bootstrap_feature: bool = False,
                   random_state: int = None) -> None:
        if base_estimator is not None: self.base_estimator = base_estimator
        if n_estimators is not None: self.n_estimators = int(n_estimators)
        if max_samples is not None: self.max_samples = float(max_samples)
        if max_features is not None: self.max_features = float(max_features)
        if bootstrap is not None: self.bootstrap = bootstrap
        if bootstrap_feature is not None: self.bootstrap_feature = bootstrap_feature
        if random_state is not None: self.random_state = random_state
    
    def __getitem__(self, index: int) -> Estimator:
        return self.estimators_[index]


class BaggingRegressor(Estimator, Estimator.Meta, Supervised):
    
    """
    A Bagging Regressor is an ensemble meta-estimator that fits base 
    regressors each on random subsets of the original dataset and then 
    averages their predictions to form the final prediction. It aims to 
    improve the stability and accuracy of machine learning algorithms, 
    especially overfitting in decision trees or other similar models.
    
    Parameters
    ----------
    `base_estimator` : Base estimator for training multiple models
    (Default `DecisionTreeRegressor`)
    `n_estimators` : Number of base estimators to fit
    `max_samples` : Maximum number of data to sample (`0~1` proportion)
    `max_features` : Maximum number of features to sample (`0~1` proporton)
    `bootstrap` : Whether to bootstrap data samples
    `bootstrap_feature`: Whether to bootstrap features
    `random_state` : Seed for random sampling
    
    Examples
    --------
    >>> bag = BaggingRegressor(base_estimator=AnyEstimator(),
                               n_estimators=100,
                               max_samples=1.0,
                               max_features=1.0,
                               bootstrap=True,
                               bootstrap_feature=False)
    >>> bag.fit(X, y)
    >>> y_pred = bag.predict(X)
    >>> est = bag[i] # Get i-th base estimator from `bag`
    
    """
    
    def __init__(self, 
                 base_estimator: Estimator = DecisionTreeRegressor,
                 n_estimators: int = 50,
                 max_samples: float | int = 1.0,
                 max_features: float | int = 1.0,
                 bootstrap: bool = True,
                 bootstrap_feature: bool = False,
                 random_state: int = None,
                 verbose: bool = False) -> None:
        self.base_estimator = base_estimator
        self.n_estimators = n_estimators
        self.max_samples = max_samples
        self.max_features = max_features
        self.bootstrap = bootstrap
        self.bootstrap_feature = bootstrap_feature
        self.random_state = random_state
        self.verbose = verbose
        self._fitted = False
    
    def fit(self, X: Matrix, y: Vector) -> 'BaggingRegressor':
        np.random.seed(self.random_state)
        self.estimators_ = []
        
        m, n = X.shape
        if self.max_samples <= 1.0: max_samples = int(m * self.max_samples)
        else: max_samples = self.max_samples
        
        if self.max_features <= 1.0: max_features = int(n * self.max_features)
        else: max_features = self.max_features
        
        for i in range(self.n_estimators):
            s_indices = np.random.choice(m, max_samples, self.bootstrap)
            f_indices = np.random.choice(n, max_features, self.bootstrap_feature)
            
            X_sample = X[s_indices][:, f_indices]
            y_sample = y[s_indices]
            
            estimator = Clone(self.base_estimator).get
            estimator.fit(X_sample, y_sample)
            self.estimators_.append((estimator, f_indices))
            
            if self.verbose:
                print(f'[Bagging] Finished fitting',
                      f'{type(self.base_estimator).__name__}',
                      f'{i + 1}/{self.n_estimators}')
        
        self._fitted = True
        return self
    
    def predict(self, X: Matrix) -> Vector:
        if not self._fitted: raise NotFittedError(self)
        
        predictions = []
        for i, (estimator, f_indices) in enumerate(self.estimators_):
            predictions.append(estimator.predict(X[:, f_indices]))
            if self.verbose:
                print(f'[Bagging] Finished prediction of',
                      f'{type(self.base_estimator).__name__}',
                      f'{i + 1}/{self.n_estimators}')
        
        return np.mean(predictions, axis=0)
    
    def score(self, X: Matrix, y: Vector, 
              metric: Evaluator = MeanSquaredError) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)
    
    def set_params(self, 
                   base_estimator: Estimator = None,
                   n_estimators: int = None,
                   max_samples: float | int = None,
                   max_features: float | int = None,
                   bootstrap: bool = False,
                   bootstrap_feature: bool = False,
                   random_state: int = None) -> None:
        if base_estimator is not None: self.base_estimator = base_estimator
        if n_estimators is not None: self.n_estimators = int(n_estimators)
        if max_samples is not None: self.max_samples = float(max_samples)
        if max_features is not None: self.max_features = float(max_features)
        if bootstrap is not None: self.bootstrap = bootstrap
        if bootstrap_feature is not None: self.bootstrap_feature = bootstrap_feature
        if random_state is not None: self.random_state = random_state
    
    def __getitem__(self, index: int) -> Estimator:
        return self.estimators_[index]




from typing import Tuple
from scipy.spatial.distance import cdist
import numpy as np

from luma.interface.util import Matrix
from luma.interface.exception import NotFittedError


__all__ = (
    'LocalOutlierFactor'
)


class LocalOutlierFactor:
    
    """
    The Local Outlier Factor (LOF) algorithm identifies outliers by comparing 
    the local density of a data point with the densities of its neighbors. 
    Points with significantly lower density than their neighbors are considered 
    outliers. LOF scores greater than 1 indicate potential outliers. It is 
    effective in datasets with varying densities and does not require a prior 
    assumption about the data distribution.
    
    Parameters
    ----------
    `n_neighbors` : Number of neighbors to estimate the local densities
    
    Examples
    --------
    >>> lof = LocalOutlierFactor()
    >>> lof.fit(X)
    
    Getting LOF Scores
    >>> lof_scores = lof.get_scores(X)
    
    Filtering Dataset
    >>> X_, y_ = lof.filter(X, y) # Supervised
    >>> X_ = lof.filter(X) # Unsupervised
    
    """
    
    def __init__(self, n_neighbors: int = 10):
        self.n_neighbors = n_neighbors
        self.lrd_ = None
        self.neighbors_ = None
        self._fitted = False
        
    def fit(self, X: Matrix) -> 'LocalOutlierFactor':
        n_samples = len(X)
        self.lrd_ = np.zeros(n_samples)
        self.neighbors_ = []

        for i in range(n_samples):
            distances = cdist([X[i]], X)[0]
            neighbors_idx = np.argsort(distances)[1:self.n_neighbors + 1]
            neighbors = X[neighbors_idx]

            self.lrd_[i] = self._local_reachability_density(X[i], neighbors)
            self.neighbors_.append(neighbors_idx)

        self._fitted = True
        return self

    def _k_distance(self, x: Matrix, neighbors: Matrix) -> Matrix:
        sorted_ = np.sort(cdist([x], neighbors)[0]) 
        return sorted_[self.n_neighbors - 1]

    def _reachability_distance(self, x: Matrix, 
                               neighbor: Matrix, 
                               neighbors: Matrix) -> Matrix:
        k_dist_neighbor = self._k_distance(neighbor, neighbors)
        return max(k_dist_neighbor, np.linalg.norm(x - neighbor))

    def _local_reachability_density(self, x: Matrix, 
                                    neighbors: Matrix) -> Matrix:
        reach_distances = [self._reachability_distance(x, neighbor, neighbors) 
                           for neighbor in neighbors]
        return 1 / (np.sum(reach_distances) / len(neighbors) 
                    if len(neighbors) > 0 else 1)

    def get_scores(self, X: Matrix) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        lof = np.zeros(len(X))
        for i in range(len(X)):
            lrd_sum = sum(self.lrd_[idx] for idx in self.neighbors_[i])
            lof[i] = lrd_sum / (self.n_neighbors * self.lrd_[i])

        return lof
    
    def filter(self, X: Matrix, y: Matrix = None,
               threshold: float = 1.5) -> Tuple[Matrix, Matrix | None]:
        lof_scores = self.get_scores(X)
        condition = lof_scores < threshold
        if y is None: return X[condition]
        else: return X[condition], y[condition]
    
    def set_params(self, n_neighbors: int = None) -> None:
        if n_neighbors is not None: self.n_neighbors = int(n_neighbors)




from typing import Tuple
import numpy as np

from luma.interface.util import Matrix, Vector


__all__ = (
    'TrainTestSplit'
)


class TrainTestSplit:
    
    """
    Splits the original dataset into the train set and the test set.
    
    Parameters
    ----------
    `X` : Feature data
    `y` : Target data (as a 1-D `Vector`)
    `test_size` : Proportional size of the test set (e.g. `0.2`, `0.3`)
    `random_state` : Seed for random sampling for split
    
    Properties
    ----------
    `get` : Returns the split data as a 4-tuple
    
    Examples
    --------
    >>> X_train, X_test, y_train, y_test = TrainTestSplit(X, y, ...).get
    
    """
    
    def __init__(self, 
                 X: Matrix,
                 y: Vector,
                 test_size: float = 0.3,
                 random_state: int = None) -> None:
        self.X = X
        self.y = y
        self.test_size = test_size
        self.random_state = random_state
    
    @property
    def get(self) -> Tuple[Matrix, Matrix, Vector, Vector]:
        if self.X.shape[0] != self.y.shape[0]:
            raise ValueError("Sample size mismatch between 'X' and 'y'!")
        
        num_samples = self.X.shape[0]
        num_test_samples = int(self.test_size * num_samples)
        indices = np.arange(num_samples)
        
        if self.random_state is not None:
            np.random.seed(self.random_state)
        np.random.shuffle(indices)

        test_indices = indices[:num_test_samples]
        train_indices = indices[num_test_samples:]

        X_train = self.X[train_indices]
        X_test = self.X[test_indices]
        y_train = self.y[train_indices]
        y_test = self.y[test_indices]

        return X_train, X_test, y_train, y_test




from typing import Tuple
import numpy as np

from luma.interface.super import Estimator, Evaluator, Supervised
from luma.interface.util import Matrix, Vector
from luma.interface.exception import NotFittedError
from luma.metric.regression import MeanSquaredError
from luma.regressor.linear import LinearRegressor


__all__ = (
    'RANSAC',
    'MLESAC'
)


class RANSAC(Estimator, Estimator.Meta, Supervised):
    
    """
    RANSAC (Random Sample Consensus) is an iterative algorithm used for robust 
    estimation of parameters from a set of data containing outliers. It works by 
    randomly selecting a subset of the original data and fitting a model to this 
    subset. The algorithm then tests all other data points against this model, 
    classifying them as inliers or outliers based on a predefined threshold. 
    This process is repeated multiple times, each time selecting a different 
    subset. The best model is chosen based on the maximum number of inliers 
    it includes.
    
    Parameters
    ----------
    `estimator` : Regression estimator (Default `LinearRegressor()`)
    `min_points` : Mininum sample size for each random sample
    `max_iter` : Maximum iteration
    `min_inliers` : Minimum number of inliers for a model to be considered valid
    `threshold` : Maximum distance a point can have to be considered an inlier
    `random_state` : Seed for random sampling the data
    
    Properties
    ----------
    Get best estimator:
    ```py
        @property
        def best_estimator(self) -> Estimator
    ```
    """
    
    def __init__(self, 
                 estimator: Estimator = LinearRegressor(),
                 min_points: int | float = 2,
                 max_iter: int = 1000,
                 min_inliers: int | float = 0.5,
                 threshold: float = None,
                 random_state: int = None,
                 verbose: bool = False) -> None:
        self.estimator = estimator
        self.min_points = min_points
        self.max_iter = max_iter
        self.min_inliers = min_inliers
        self.threshold = threshold
        self.random_state = random_state
        self.verbose = verbose
        self.inliers_ = None
        self._fitted = False
    
    def fit(self, X: Matrix, y: Vector) -> 'RANSAC':
        m, _ = X.shape
        best_err = np.inf
        best_indices = None
        
        np.random.seed(self.random_state)
        Xy = np.hstack((X, y.reshape(-1, 1)))
        
        if self.min_points < 1: self.min_points = np.ceil(m * self.min_points)
        if self.min_inliers < 1: self.min_inliers = np.ceil(m * self.min_inliers)
        if self.threshold is None: self.threshold = self._mean_absdev(y)
        
        for i in range(self.max_iter):
            hypo_indices, test_indices = self._random_partition(X)
            hypo_inliers = Xy[hypo_indices]
            test_points = Xy[test_indices]
            
            self.estimator.fit(hypo_inliers[:, :-1], hypo_inliers[:, -1])
            test_err = self._error(test_points[:, :-1], test_points[:, -1])
            
            true_indices = test_indices[test_err < self.threshold]
            true_inliers = Xy[true_indices]
            
            if len(true_inliers) > self.min_inliers:
                Xy_better = np.concatenate((hypo_inliers, true_inliers))
                
                self.estimator.fit(Xy_better[:, :-1], Xy_better[:, -1])
                better_err = self._error(Xy_better[:, :-1], Xy_better[:, -1])
                this_err = np.mean(better_err)
                
                if this_err < best_err:
                    best_err = this_err
                    best_indices = np.concatenate((hypo_indices, true_indices))
            
            if self.verbose and i % 100 == 0:
                print(f'[RANSAC] Iteration {i}/{self.max_iter}',
                      f'with current best-error: {best_err}')
        
        self.inliers_ = Xy[best_indices, :-1]
        self._fitted = True
        return self
    
    def _random_partition(self, X: Matrix) -> Tuple[Vector, Vector]:
        all_indices = np.arange(X.shape[0])
        np.random.shuffle(all_indices)
        indices_l = all_indices[:self.min_points]
        indices_r = all_indices[self.min_points:]
        
        return indices_l, indices_r

    def _error(self, X: Matrix, y: Vector) -> Vector:
        y_pred = self.estimator.predict(X)
        return np.abs(y - y_pred)
    
    def _mean_absdev(self, y: Vector) -> float:
        y_median = np.median(y)
        abs_dev = np.abs(y - y_median)
        mad = np.median(abs_dev)
        return mad
    
    @property
    def best_estimator(self) -> Estimator:
        if not self._fitted: raise NotFittedError(self)
        return self.estimator
    
    def predict(self, X: Matrix) -> Vector:
        best = self.best_estimator
        return best.predict(X)

    def score(self, X: Matrix, y: Matrix, 
              metric: Evaluator = MeanSquaredError) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)
    
    def set_params(self, 
                   estimator: Estimator = None,
                   min_points: int | float = None,
                   max_iter: int = None,
                   min_inliers: int | float = None,
                   threshold: float = None,
                   random_state: int = None) -> None:
        if estimator is not None: self.estimator = estimator
        if min_points is not None: self.min_points = float(min_points)
        if max_iter is not None: self.max_iter = int(max_iter)
        if min_inliers is not None: self.min_inliers = float(min_inliers)
        if threshold is not None: self.threshold = float(threshold)
        if random_state is not None: self.random_state = int(random_state)


class MLESAC(Estimator, Estimator.Meta, Supervised):
    
    """
    MLESAC (Maximum Likelihood Estimation SAmple Consensus) is an iterative algorithm 
    used for robust estimation of parameters from a set of data containing outliers. 
    Unlike RANSAC which focuses on maximizing the number of inliers, MLESAC seeks 
    to maximize the likelihood of the model given the data, considering both the 
    number of inliers and the quality of the fit.
    
    Parameters
    ----------
    `estimator` : Regression estimator (Default `LinearRegressor()`)
    `min_points` : Minimum sample size for each random sample
    `max_iter` : Maximum number of iterations
    `min_inliers` : Minimum number of inliers for a model to be considered valid
    `threshold` : Maximum distance a point can have to be considered an inlier
    `random_state` : Seed for random sampling the data
    
    Properties
    ----------
    Get best estimator:
    ```py
        @property
        def best_estimator(self) -> Estimator
    ```
    
    Notes
    -----
    * `MLESAC` assumes that inliers follow a multivariate Gaussian model
    
    """
    
    def __init__(self, 
                 estimator: Estimator = LinearRegressor(),
                 min_points: int | float = 2,
                 max_iter: int = 1000,
                 min_inliers: int | float = 0.5,
                 threshold: float = None,
                 random_state: int = None,
                 verbose: bool = False) -> None:
        self.estimator = estimator
        self.min_points = min_points
        self.max_iter = max_iter
        self.min_inliers = min_inliers
        self.threshold = threshold
        self.random_state = random_state
        self.verbose = verbose
        self.inliers_ = None
        self._fitted = False
    
    def fit(self, X: Matrix, y: Vector) -> 'MLESAC':
        m, _ = X.shape
        best_likelihood = -np.inf
        best_indices = None
        
        np.random.seed(self.random_state)
        Xy = np.hstack((X, y.reshape(-1, 1)))
        
        if self.min_points < 1: self.min_points = np.ceil(m * self.min_points)
        if self.min_inliers < 1: self.min_inliers = np.ceil(m * self.min_inliers)
        if self.threshold is None: self.threshold = self._mean_absdev(y)
        
        for i in range(self.max_iter):
            hypo_indices, test_indices = self._random_partition(X)
            hypo_inliers = Xy[hypo_indices]
            test_points = Xy[test_indices]
            
            self.estimator.fit(hypo_inliers[:, :-1], hypo_inliers[:, -1])
            test_err = self._error(test_points[:, :-1], test_points[:, -1])
            
            true_indices = test_indices[test_err < self.threshold]
            true_inliers = Xy[true_indices]
            
            if len(true_inliers) >= self.min_inliers:
                Xy_better = np.concatenate((hypo_inliers, true_inliers))
                
                self.estimator.fit(Xy_better[:, :-1], Xy_better[:, -1])
                better_err = self._error(Xy_better[:, :-1], Xy_better[:, -1])
                likelihood = self._calculate_likelihood(better_err)
                
                if likelihood > best_likelihood:
                    best_likelihood = likelihood
                    best_indices = np.concatenate((hypo_indices, true_indices))

            if self.verbose and i % 100 == 0:
                print(f'[MLESAC] Iteration {i}/{self.max_iter}',
                      f'with current best-likelihood: {best_likelihood}')

        self.inliers_ = Xy[best_indices, :-1]
        self._fitted = True
        return self
    
    def _random_partition(self, X: Matrix) -> Tuple[Vector, Vector]:
        all_indices = np.arange(X.shape[0])
        np.random.shuffle(all_indices)
        indices_l = all_indices[:self.min_points]
        indices_r = all_indices[self.min_points:]
        
        return indices_l, indices_r

    def _error(self, X: Matrix, y: Vector) -> Vector:
        y_pred = self.estimator.predict(X)
        return np.abs(y - y_pred)
    
    def _mean_absdev(self, y: Vector) -> float:
        y_median = np.median(y)
        abs_dev = np.abs(y - y_median)
        mad = np.median(abs_dev)
        return mad
    
    def _calculate_likelihood(self, error: Vector) -> float:
        if error.ndim == 1: error = error[:, np.newaxis]
        m, n = error.shape
        cov_matrix = np.cov(error, rowvar=False)
        
        if n == 1: cov_matrix = cov_matrix.reshape(1, 1)
        det_cov_matrix = np.linalg.det(cov_matrix)
        inv_cov_matrix = np.linalg.inv(cov_matrix)

        likelihood = -m * n * np.log(2 * np.pi) / 2
        likelihood -= m * np.log(det_cov_matrix) / 2
        for error in error:
            likelihood -= 0.5 * np.dot(error, np.dot(inv_cov_matrix, error))

        return likelihood
    
    @property
    def best_estimator(self) -> Estimator:
        if not self._fitted: raise NotFittedError(self)
        return self.estimator
    
    def predict(self, X: Matrix) -> Vector:
        best = self.best_estimator
        return best.predict(X)

    def score(self, X: Matrix, y: Matrix, 
              metric: Evaluator = MeanSquaredError) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)
    
    def set_params(self, 
                   estimator: Estimator = None,
                   min_points: int | float = None,
                   max_iter: int = None,
                   min_inliers: int | float = None,
                   threshold: float = None,
                   random_state: int = None) -> None:
        if estimator is not None: self.estimator = estimator
        if min_points is not None: self.min_points = float(min_points)
        if max_iter is not None: self.max_iter = int(max_iter)
        if min_inliers is not None: self.min_inliers = float(min_inliers)
        if threshold is not None: self.threshold = float(threshold)
        if random_state is not None: self.random_state = int(random_state)




from scipy.spatial import distance_matrix
import numpy as np

from luma.interface.util import Matrix
from luma.interface.super import Estimator, Supervised, Evaluator
from luma.interface.util import NearestNeighbors
from luma.interface.exception import NotFittedError
from luma.metric.classification import Accuracy


__all__ = (
    'KNNClassifier', 
    'AdaptiveKNNClassifier', 
    'WeightedKNNClassifier'
)


class KNNClassifier(Estimator, Supervised):
    
    """
    The K-Nearest Neighbors (KNN) classifier is a simple and intuitive 
    machine learning algorithm that classifies a new data point based on 
    the majority class among its closest neighbors. It operates by 
    calculating the distance (such as Euclidean distance) between the 
    new point and all points in the training set, identifying the 'k' 
    nearest neighbors, and then performing a majority vote among these 
    neighbors to determine the class label.
    
    Parameters
    ----------
    `n_neighbors` : Number of neighbors to be considered close
    
    """
    
    def __init__(self, n_neighbors: int = 5):
        self.n_neighbors = n_neighbors
        self._neighbors = None
        self._y = None
        self._fitted = False

    def fit(self, X: Matrix, y: Matrix) -> 'KNNClassifier':
        self._neighbors = NearestNeighbors(X, self.n_neighbors)
        self._y = y
        self._fitted = True
        return self

    def predict(self, X: Matrix) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        predictions = []
        
        for x in X:
            distances = np.linalg.norm(self._neighbors.data - x, axis=1)
            nearest_neighbor_ids = np.argsort(distances)[:self.n_neighbors]
            nearest_neighbor_labels = self._y[nearest_neighbor_ids]
            most_common_label = np.bincount(nearest_neighbor_labels).argmax()
            predictions.append(most_common_label)

        return np.array(predictions)

    def score(self, X: Matrix, y: Matrix, 
              metric: Evaluator = Accuracy) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)
    
    def set_params(self, n_neighbors: int = None) -> None:
        if n_neighbors is not None: self.n_neighbors = int(n_neighbors)


class AdaptiveKNNClassifier(Estimator, Supervised):
    
    """
    The Adaptive K-Nearest Neighbors (AdaKNN) Classifier is an extension 
    of the conventional KNN algorithm for classification. It classifies a 
    new data point based on the majority class among its neighbors, where 
    the number of neighbors (k) is adaptively determined based on the local 
    density of the training data. This adaptive approach allows the algorithm 
    to be more flexible and effective, especially in datasets with varying 
    densities.
    
    Parameters
    ----------
    `n_density` : Number of nearest neighbors to estimate the local density
    `min_neighbors` : Minimum number of neighbors to be considered for averaging
    `max_neighbors` : Maximum number of neighbors to be considered
    
    """
    
    def __init__(self, 
                 n_density: int = 10, 
                 min_neighbors: int = 5, 
                 max_neighbors: int = 20):
        self.n_density = n_density
        self.min_neighbors = min_neighbors
        self.max_neighbors = max_neighbors
        self._X = None
        self._y = None
        self._fitted = False

    def fit(self, X: Matrix, y: Matrix) -> 'AdaptiveKNNClassifier':
        self._X = X
        self._y = y
        
        self.dist_matrix = distance_matrix(X, X)
        self.local_density = np.sort(self.dist_matrix, axis=1)
        self.local_density = self.local_density[:, self.n_density]
        
        self.k_values = np.clip(np.ceil(self.max_neighbors / self.local_density), 
                           self.min_neighbors, self.max_neighbors).astype(int)
        
        self._fitted = True
        return self

    def predict(self, X: Matrix) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        dist_matrix = distance_matrix(X, self._X)
        
        predictions = []
        for point_distances in dist_matrix:
            nearest_indices = np.argsort(point_distances)
            votes = np.zeros(len(np.unique(self._y)))

            for idx, _ in enumerate(point_distances[nearest_indices]):
                if idx >= self.k_values[nearest_indices[idx]]: break
                votes[self._y[nearest_indices[idx]]] += 1

            most_common = np.argmax(votes)
            predictions.append(most_common)

        return np.array(predictions)

    def score(self, X: Matrix, y: Matrix, 
              metric: Evaluator = Accuracy) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)
    
    def set_params(self, 
                   n_density: int = None,
                   min_neighbors: int = None,
                   max_neighbors: int = None) -> None:
        if n_density is not None: self.n_density = int(n_density)
        if min_neighbors is not None: self.min_neighbors = int(min_neighbors)
        if max_neighbors is not None: self.max_neighbors = int(max_neighbors)


class WeightedKNNClassifier(Estimator, Supervised):
    
    """
    The Weighted KNN Classifier is a variation of the k-Nearest Neighbors algorithm 
    where neighbors contribute to the classification decision based on their distance 
    from the query point. Closer neighbors have more influence as they are assigned 
    higher weights, typically using inverse distance weighting. This approach enhances 
    prediction accuracy, especially in unevenly distributed datasets.
    
    Parameters
    ----------
    `n_neighbors` : Number of neighbors to be considered close
    
    """
    
    def __init__(self, n_neighbors: int = 5):
        self.n_neighbors = n_neighbors
        self._X = None
        self._y = None
        self._fitted = False

    def fit(self, X: Matrix, y: Matrix) -> 'WeightedKNNClassifier':
        self._X = X
        self._y = y
        
        self._fitted = True
        return self

    def predict(self, X: Matrix) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        predictions = np.zeros(X.shape[0])

        for i, x in enumerate(X):
            distances = np.linalg.norm(self._X - x, axis=1)
            nearest_neighbors = np.argsort(distances)[:self.n_neighbors]

            weights = 1 / (distances[nearest_neighbors] + 1e-5)
            weighted_votes = np.zeros(np.unique(self._y).shape[0])
            for idx, neighbor in enumerate(nearest_neighbors):
                weighted_votes[self._y[neighbor]] += weights[idx]

            predictions[i] = np.argmax(weighted_votes)

        return predictions
    
    def score(self, X: Matrix, y: Matrix, 
              metric: Evaluator = Accuracy) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)
    
    def set_params(self, n_neighbors: int = None) -> None:
        if n_neighbors is not None: self.n_neighbors = int(n_neighbors)




"""
`LUMA`
======

Lumerico's Universal Machine-learning Assistant
-----------------------------------------------
LUMA is a comprehensive, user-friendly Python module designed for both beginners 
and advanced users in the field of machine learning and data science. It provides 
a wide range of tools and functionalities to streamline the process of data analysis, 
model building, evaluation, and deployment.

Key Features
------------
- Easy Data Handling: Simplify data preprocessing, transformation, and visualization.
- Model Building: Access a variety of machine learning algorithms.
- Model Evaluation: Utilize robust tools for model validation and tuning.

Getting Started
---------------
To get started with LUMA, install the package using `pip`:

```sh
pip install luma-ml
```

After installation, import LUMA in your Python script to access its features:

```py
import luma
```

For detailed documentation, tutorials, and examples, please refer to:
- GitHub: https://github.com/ChanLumerico/LUMA
- Notion: https://lumerico284.notion.site/LUMA
- PyPI: https://pypi.org/project/luma-ml

Contribute
----------
LUMA is an open-source project, and we welcome contributions from the community. 
Whether you're interested in fixing bugs, adding new features, or improving documentation, 
your help is appreciated.

License
-------
LUMA is released under the GPL-3.0 License. See `LICENSE` file for more details.

"""


from typing import *
from scipy.linalg import eigh
import numpy as np

from luma.interface.util import Matrix
from luma.interface.super import Transformer, Unsupervised
from luma.interface.exception import NotFittedError


__all__ = (
    'KernelPCA'
)
    

class KernelPCA(Transformer, Unsupervised):
    
    """
    Kernel PCA is a dimensionality reduction technique that extends 
    Principal Component Analysis (PCA) to capture complex, nonlinear 
    relationships in data by using a kernel function to transform 
    the data into a higher-dimensional space where nonlinear patterns 
    can be better captured.
    
    Parameters
    ----------
    `n_components` : Number of principal components
    `degree` : Polynomial degree of `poly` kernel
    `gamma` : Shape parameter of `rbf`, `sigmoid`, `laplacian`
    `coef` : Additional coefficient of `poly`, `sigmoid`
    `kernel` : Type of kernel functions
    (e.g. `linear`, `poly`, `rbf`, `sigmoid`, `laplacian`)
    
    """
    
    def __init__(self, 
                 n_components: int = None,
                 degree: int = 3,
                 gamma: float = 15.0,
                 coef: float = 1.0,
                 kernel: Literal['linear', 'poly', 'rbf', 
                                 'sigmoid', 'laplacian'] = 'linear') -> None:
        self.n_components = n_components
        self.degree = degree
        self.gamma = gamma
        self.coef = coef
        self.kernel = kernel
        self.X = None
        self._fitted = False
    
    def fit(self, X: Matrix) -> 'KernelPCA':
        self.X = X
        if self.kernel == 'linear': self.kernel_func = self._linear
        elif self.kernel == 'poly': self.kernel_func = self._poly
        elif self.kernel == 'rbf': self.kernel_func = self._rbf
        elif self.kernel == 'sigmoid': self.kernel_func = self._sigmoid
        elif self.kernel == 'laplacian': self.kernel_func = self._laplacian
        else: raise ValueError('[KPCA] Unsupported Kernel!')

        N = X.shape[0]
        self.K = np.zeros((N, N))
        for i in range(N):
            for j in range(i, N):
                kernel_value = self.kernel_func(X[i], X[j])
                self.K[i, j] = kernel_value
                self.K[j, i] = kernel_value
        
        one_n = np.ones((N, N)) / N
        self.K = self.K - one_n.dot(self.K) - self.K.dot(one_n)
        self.K += one_n.dot(self.K).dot(one_n)
        self.eigvals, self.eigvecs = eigh(self.K)
        self.eigvals, self.eigvecs = self.eigvals[::-1], self.eigvecs[:, ::-1]
        self._fitted =  True
        return self

    def transform(self) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        return np.column_stack([self.eigvecs[:, i] for i in range(self.n_components)])
        
    def fit_transform(self, X: Matrix) -> Matrix:
        self.fit(X)
        return self.transform()
    
    def _linear(self, x: Matrix, y: Matrix) -> Matrix:
        return np.dot(x, y)
    
    def _poly(self, x: Matrix, y: Matrix) -> Matrix:
        return (np.dot(x, y) + self.coef) ** self.degree
    
    def _rbf(self, x: Matrix, y: Matrix) -> Matrix:
        if self.gamma is None: self.gamma = 1 / self.X.shape[1]
        return np.exp(-self.gamma * np.linalg.norm(x - y) ** 2)
    
    def _sigmoid(self, x: Matrix, y: Matrix) -> Matrix:
        return np.tanh(self.gamma * np.dot(x, y) + self.coef)
    
    def _laplacian(self, x: Matrix, y: Matrix) -> Matrix:
        return np.exp(-self.gamma * np.linalg.norm(x - y))
    
    def set_params(self,
                   n_components: int = None,
                   degree: int = None,
                   gamma: float = None,
                   coef: float = None,
                   kernel: Literal = None) -> None:
        if n_components is not None: self.n_components = int(n_components)
        if degree is not None: self.degree = int(degree)
        if gamma is not None: self.gamma = float(gamma)
        if coef is not None: self.coef = float(coef)
        if kernel is not None: self.kernel = str(kernel)




import numpy as np

from luma.interface.util import Matrix
from luma.interface.exception import NotFittedError
from luma.interface.super import Estimator, Evaluator, Supervised
from luma.metric.regression import MeanSquaredError


__all__ = (
    'LinearRegressor', 
    'RidgeRegressor', 
    'LassoRegressor', 
    'ElasticNetRegressor'
)


class RidgeRegressor(Estimator, Supervised):
    
    """
    Ridge regression is a linear regression technique used to 
    prevent overfitting in predictive models. 
    It adds a penalty term called "L2 regularization" to help 
    reduce the complexity of the model and prevent it from 
    fitting noise in the data.
    
    Parameters
    ----------
    `alpha` : L2-regularization strength
    
    """
    
    def __init__(self, alpha: float = 1.0) -> None:
        self.alpha = alpha
        self._fitted = False

    def fit(self, X: Matrix, y: Matrix) -> 'RidgeRegressor':
        X = np.column_stack((np.ones(X.shape[0]), X))
        identity_matrix = np.identity(X.shape[1])
        self.coefficients = np.linalg.inv(X.T.dot(X) + self.alpha * identity_matrix)
        self.coefficients = self.coefficients.dot(X.T).dot(y)
        self._fitted = True
        return self

    def predict(self, X: Matrix) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        X = np.column_stack((np.ones(X.shape[0]), X))
        return X.dot(self.coefficients)
    
    def score(self, X: Matrix, y: Matrix, 
              metric: Evaluator = MeanSquaredError) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)

    def set_params(self, alpha: float = None) -> None:
        if alpha is not None: self.alpha = float(alpha)


class LassoRegressor(Estimator, Supervised):
    
    """
    Lasso regression is a linear regression technique used for 
    feature selection and regularization. It adds a penalty term, 
    called "L1 regularization," to the standard linear regression 
    objective function. Lasso encourages some of the model's 
    coefficients to become exactly zero, effectively eliminating 
    certain features from the model.
    
    Parameters
    ----------
    `alpha` : L1-regularization strength
    `max_iter` : Number of iteration
    `learning_rate` : Step size of the gradient descent update
    
    """
    
    def __init__(self, 
                 alpha: float = 1.0, 
                 max_iter: int = 100, 
                 learning_rate: float = 0.01, 
                 verbose: bool = False) -> None:
        self.alpha = alpha
        self.max_iter = max_iter
        self.learning_rate = learning_rate
        self.verbose = verbose
        self._fitted = False

    def _soft_threshold(self, x: Matrix, threshold: float) -> Matrix:
        return np.sign(x) * np.maximum(0, np.abs(x) - threshold)

    def fit(self, X: Matrix, y: Matrix) -> 'LassoRegressor':
        X = np.column_stack((np.ones(X.shape[0]), X))
        self.coefficients = np.zeros(X.shape[1])
        
        for i in range(self.max_iter):
            coefficients_prev = self.coefficients.copy()
            y_pred = X.dot(self.coefficients)
            gradient = -(X.T.dot(y - y_pred)) * self.learning_rate
            self.coefficients = self.coefficients - (1.0 / X.shape[0]) * gradient
            self.coefficients = self._soft_threshold(self.coefficients, self.alpha / X.shape[0])
            
            if self.verbose and i % 10 == 0:
                print(f'[LassoReg] iteration: {i}/{self.max_iter}', end='')
                print(f' - delta-coeff norm: {np.linalg.norm(self.coefficients - coefficients_prev)}')
        
        self._fitted = True
        return self

    def predict(self, X: Matrix) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        X = np.column_stack((np.ones(X.shape[0]), X))
        return X.dot(self.coefficients)
    
    def score(self, X: Matrix, y: Matrix, 
              metric: Evaluator = MeanSquaredError) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)
    
    def set_params(self, 
                   alpha: float = None, 
                   max_iter: int = None,
                   learning_rate: float = None) -> None:
        if alpha is not None: self.alpha = float(alpha)
        if max_iter is not None: self.max_iter = int(max_iter)
        if learning_rate is not None: self.learning_rate = float(learning_rate)


class ElasticNetRegressor(Estimator, Supervised):
    
    """
    Elastic-Net regression is a linear regression technique 
    that combines both L1 (Lasso) and L2 (Ridge) regularization methods. 
    It adds a combination of L1 and L2 penalty terms to the 
    standard linear regression objective function.
    
    Parameters
    ----------
    `alpha` : Regularization strength
    `rho` : Balancing parameter between `l1` and `l2`
    `max_iter` : Number of iteration
    `learning_rate` : Step size of the gradient descent update
    
    """
    
    def __init__(self, 
                 alpha: float = 1.0, 
                 rho: float = 0.5, 
                 max_iter: int = 100, 
                 learning_rate: float = 0.01,
                 verbose: bool = False) -> None:
        self.alpha = alpha
        self.rho = rho
        self.max_iter = max_iter
        self.learning_rate = learning_rate
        self.verbose = verbose
        self.coef_ = None
        self._fitted = False
        
    def _soft_threshold(self, x: Matrix, alpha: float) -> Matrix:
        return np.sign(x) * np.maximum(np.abs(x) - alpha, 0)

    def fit(self, X: Matrix, y: Matrix) -> 'ElasticNetRegressor':
        N, p = X.shape
        self.coef_ = np.zeros(p)

        for i in range(self.max_iter):
            y_pred = X.dot(self.coef_)
            gradient = -(1 / N) * X.T.dot(y - y_pred)
            lefthand = self.coef_ - self.learning_rate * gradient
            righthand = self.alpha * self.rho
            
            self.coef_ = self._soft_threshold(lefthand, righthand)
            self.coef_ /= 1 + self.alpha * (1 - self.rho)
            
            if self.verbose and i % 10 == 0: 
                print(f'[ElasticReg] iteration: {i}/{self.max_iter}')
        
        self._fitted = True
        return self

    def predict(self, X: Matrix) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        if self.coef_ is None: raise ValueError()
        return X.dot(self.coef_)
    
    def score(self, X: Matrix, y: Matrix, 
              metric: Evaluator = MeanSquaredError) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)
    
    def set_params(self,
                   alpha: float = None,
                   rho: float = None,
                   max_iter: int = None,
                   learning_rate: float = None) -> None:
        if alpha is not None: self.alpha = float(alpha)
        if rho is not None: self.rho = float(rho)
        if max_iter is not None: self.max_iter = int(max_iter)
        if learning_rate is not None: self.learning_rate = float(learning_rate)


class LinearRegressor(Estimator, Supervised):
    
    """
    An Ordinary Least Squares (OLS) Linear Regressor is a statistical method 
    used in linear regression analysis. It estimates the coefficients of the 
    linear equation, minimizing the sum of the squared differences between 
    observed and predicted values. This results in a line of best fit through 
    the data points in multidimensional space. OLS is widely used for its 
    simplicity and efficiency in modeling linear relationships.
    """
    
    def __init__(self):
        self.coefficients = None
        self._fitted = False

    def fit(self, X: Matrix, y: Matrix) -> 'LinearRegressor':
        X = np.hstack([np.ones((X.shape[0], 1)), X])
        X_T = np.transpose(X)
        self.coefficients = np.linalg.inv(X_T.dot(X)).dot(X_T).dot(y)
        
        self._fitted = True
        return self

    def predict(self, X: Matrix) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        X = np.hstack([np.ones((X.shape[0], 1)), X])
        
        return X.dot(self.coefficients)
    
    def score(self, X: Matrix, y: Matrix, 
              metric: Evaluator = MeanSquaredError) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)

    def set_params(self) -> None: ...




from typing import Optional
from matplotlib import colors, pyplot as plt
import numpy as np

from luma.interface.super import Visualizer
from luma.interface.util import Matrix


__all__ = (
    'GraphPlot'
)


class GraphPlot(Visualizer):
    
    """
    A visualizer which plots the structure of the given graph.
    For a weighted graph, edge weights are represented as a 
    gradient color of a colormap, while the edges of an
    unweighted graph are shown as an uniform color.
    
    Parameters
    ----------
    `nodes` : Coordinate matrix of nodes
    `edges` : Adjacency matrix of a graph
    
    Notes
    -----
    * Currently, only adjacency matrices are supported.
    * Unweighted graphs are not supported temporarily.
    
    """
    
    def __init__(self,
                 nodes: Matrix,
                 edges: Matrix,
                 xlabel: str = r'$x_1$',
                 ylabel: str = r'$x_2$',
                 title: str | None = None,
                 grid: bool = False) -> None:
        self.nodes = nodes
        self.edges = edges
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.title = title
        self.grid = grid
    
    def plot(self, 
             ax: Optional[plt.Axes] = None, 
             size: float = 250, 
             scale: float = 10.0,
             show: bool = False) -> plt.Axes:
        x_min, x_max = self.nodes[:, 0].min(), self.nodes[:, 0].max()
        y_min, y_max = self.nodes[:, 1].min(), self.nodes[:, 1].max()
        delta_x, delta_y = (x_max - x_min) / size, (y_max - y_min) / size
        
        m, _ = self.nodes.shape
        i_upper, j_upper = np.triu_indices(m, k=1)
        weights = self.edges[i_upper, j_upper]
        valid_edges = (weights != 0) & (weights != np.inf)

        edges = np.column_stack((i_upper[valid_edges], j_upper[valid_edges]))
        weights = weights[valid_edges]

        norm = colors.Normalize(vmin=weights.min(), vmax=weights.max())
        cmap = plt.cm.plasma

        if ax is None: _, ax = plt.subplots()
        ax.scatter(self.nodes[:, 0], self.nodes[:, 1], c='black', s=20)

        for edge, weight in zip(edges, weights):
            points = self.nodes[edge]
            ax.plot(points[:, 0], points[:, 1], 
                    color=cmap(norm(weight)), linewidth=2, alpha=0.8)

        sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
        sm.set_array([])
        plt.colorbar(sm, ax=ax, label='Edge Weights')

        ax.set_aspect('equal')
        ax.set_xlim(x_min - delta_x * scale, x_max + delta_x * scale)
        ax.set_ylim(y_min - delta_y * scale, y_max + delta_y * scale)
        
        ax.set_xlabel(self.xlabel)
        ax.set_ylabel(self.ylabel)
        ax.set_title('Graph Plot' if self.title is None else self.title)
        ax.figure.tight_layout()
        
        if self.grid: ax.grid()
        if show: plt.show()
        return ax




from typing import Optional
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd

from luma.interface.super import Visualizer


__all__ = (
    'CorrelationHeatMap', 
    'CorrelationBar', 
    'JointPlot', 
    'MissingProportion'
)


class CorrelationHeatMap(Visualizer):
    def __init__(self, data: pd.DataFrame) -> None:
        self.data = data
        self.corr = data.corr()
    
    def plot(self, 
             ax: Optional[plt.Axes] = None, 
             colorMap: str = 'rocket', 
             annotate: bool = True, 
             colorBar: bool = True,
             show: bool = False) -> plt.Axes:
        if ax is None:
            n_features = self.data.shape[1]
            size = n_features / 2 if n_features < 20 else 10
            _, ax = plt.subplots(figsize=(size + 1, size))

        sns.heatmap(self.corr, 
                    ax=ax, 
                    cmap=colorMap, 
                    annot=annotate, 
                    cbar=colorBar, 
                    fmt='0.2f')
        
        ax.set_title('Correlation Heat Map')
        ax.figure.tight_layout()
        
        if show: plt.show()
        return ax


class CorrelationBar(Visualizer):
    def __init__(self, data: pd.DataFrame, target: str) -> None:
        self.data = data
        self.target = target
    
    def plot(self, 
             ax: Optional[plt.Axes] = None, 
             show: bool = False) -> plt.Axes:
        if ax is None: _, ax = plt.subplots()
        corr_bar = [abs(self.data[col].corr(self.data[self.target])) for col in self.data]
        sns.barplot(x=self.data.columns, y=corr_bar, hue=self.data.columns, ax=ax)
        
        ax.set_xticklabels(ax.get_xticklabels(), rotation=45)
        ax.set_title(f'Correlations with {self.target}')
        ax.set_xlabel('Features')
        ax.set_ylabel('Correlation')
        ax.figure.tight_layout()

        if show: plt.show()
        return ax


class JointPlot(Visualizer):
    def __init__(self, data: pd.DataFrame, x: str, y: str) -> None:
        self.data = data
        self.x = x
        self.y = y

    def plot(self, color: str = 'tab:blue') -> None:    
        sns.jointplot(data=self.data, x=self.x, y=self.y, kind='reg', color=color)
        plt.title(f'{self.x} vs. {self.y}')
        plt.tight_layout()
        plt.show()


class MissingProportion(Visualizer):
    def __init__(self, data: pd.DataFrame):
        self.data = data

    def nan_proportions(self) -> pd.DataFrame:
        nan_props = self.data.isna().mean()
        return nan_props

    def plot(self, 
             ax: Optional[plt.Axes] = None,
             show: bool = False) -> plt.Axes:
        if ax is None: _, ax = plt.subplots()
        nan_props = self.nan_proportions()
        sns.barplot(x=nan_props.index, y=nan_props.values, hue=self.data.columns, ax=ax)
        
        ax.set_xticklabels(ax.get_xticklabels(), rotation=45)
        ax.set_xlabel('Columns')
        ax.set_ylabel('Proportion')
        ax.set_title('Missing Value Proportions')
        ax.figure.tight_layout()
        
        if show: plt.show()
        return ax




from typing import List, Literal
import numpy as np

from luma.interface.super import Estimator, Evaluator, Supervised
from luma.interface.util import Matrix, Vector, Scalar, Clone
from luma.interface.exception import NotFittedError, UnsupportedParameterError
from luma.metric.classification import Accuracy
from luma.metric.regression import MeanSquaredError
from luma.preprocessing.encoder import LabelEncoder


__all__ = (
    'VotingClassifier',
    'VotingRegressor'
)


class VotingClassifier(Estimator, Estimator.Meta, Supervised):
    
    """
    A Voting Classifier is an ensemble machine learning model that combines 
    predictions from multiple base models. It uses a voting mechanism, either 
    'hard' (majority rule) or 'soft' (weighted probabilities), to determine 
    the final output. This approach can increase prediction accuracy and 
    robustness by leveraging the strengths of diverse models. It's 
    particularly useful when individual models have different error patterns. 
    The Voting Classifier is widely used in classification tasks to improve 
    performance over single models.
    
    Parameters
    ----------
    `estimators` : List of estimators to vote
    `voting` : Voting critetion
    (`label` to vote the most frequent and `prob` to vote the most probable)
    `weights` : Weights for each classifier on voting
    (`None` for uniform weights)
    
    Examples
    --------
    >>> vot = VotingClassifier(estimators=[AnyClassifier(), ...], 
                               voting='label',
                               weights=[0.25, 0.5, ...])
    >>> vot.fit(X, y)
    >>> pred = vot.predict(X)
    >>> clf = vot[i] # Get i-th classifier from `vot`
    
    """
    
    def __init__(self,
                 estimators: List[Estimator] = None,
                 voting: Literal['label', 'prob'] = 'label',
                 weights: Vector[Scalar] = None,
                 verbose: bool = False) -> None:
        self.estimators = estimators
        self.voting = voting
        self.weights = weights
        self.verbose = verbose
        self._fitted = False
    
    def fit(self, X: Matrix, y: Vector) -> 'VotingClassifier':
        if self.voting not in ('label', 'prob'):
            raise UnsupportedParameterError(self.voting)
        
        if self.weights and len(self.weights) != len(self.estimators):
            raise ValueError(f"Size mismatch between 'weights' and 'estimators'!")
        
        self.le = LabelEncoder()
        self.le.fit(y)
        
        self.classes = self.le.classes
        self.estimators_ = []
        
        for est in self.estimators:
            fitted_est = Clone(est).get
            if hasattr(fitted_est, 'verbose'):
                fitted_est.verbose = self.verbose
            
            fitted_est.fit(X, self.le.transform(y))
            self.estimators_.append(fitted_est)
        
        self._fitted = True
        return self
    
    def predict(self, X: Matrix) -> Vector:
        if not self._fitted: raise NotFittedError(self)
        
        if self.voting == 'label':
            preds = np.asarray([est.predict(X) for est in self.estimators_]).T
            majority_vote = np.apply_along_axis(
                lambda x: np.argmax(np.bincount(x, weights=self.weights)),
                axis=1,
                arr=preds
            )
        else:
            majority_vote = np.argmax(self.predict_proba(X), axis=1)
        
        majority_vote = self.le.inverse_transform(majority_vote)
        return majority_vote
    
    def predict_proba(self, X: Matrix) -> Matrix:
        probas = []
        for est in self.estimators_:
            if not hasattr(est, 'predict_proba'):
                raise ValueError(f"'{type(est)}' does not support 'predict_proba'")
            probas.append(est.predict_proba(X))
        
        return np.average(probas, axis=0, weights=self.weights)

    def score(self, X: Matrix, y: Vector, metric: Evaluator = Accuracy) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)
    
    def set_params(self, 
                   estimators: List[Estimator] = None,
                   voting: Literal = None,
                   weights: Vector[Scalar] = None) -> None:
        if estimators is not None: self.estimators = estimators
        if voting is not None: self.voting = str(voting)
        if weights is not None: self.weights = weights

    def __getitem__(self, index: int) -> Estimator:
        return self.estimators_[index]


class VotingRegressor(Estimator, Estimator.Meta, Supervised):
    
    """
    A Voting Regressor is an ensemble machine learning model that combines 
    predictions from multiple base regression models. It uses a voting mechanism, 
    where the final output is the average of the predictions from all models. 
    This approach can increase prediction accuracy and robustness by leveraging 
    the strengths of diverse models. It's particularly useful when individual 
    models have different error patterns. The Voting Regressor is widely used in 
    regression tasks to improve performance over single models.
    
    Parameters
    ----------
    `estimators` : List of regressors to vote
    `weights` : Weights for each regressor on voting
    (`None` for uniform weights)
    
    Examples
    --------
    >>> vot = VotingRegressor(estimators=[AnyRegressor(), ...], 
                              weights=[0.25, 0.5, ...])
    >>> vot.fit(X, y)
    >>> pred = vot.predict(X)
    >>> reg = vot[i] # Get i-th regressor from `vot`
    
    """
    
    def __init__(self,
                 estimators: List[Estimator] = None,
                 weights: Vector[Scalar] = None,
                 verbose: bool = False) -> None:
        self.estimators = estimators
        self.weights = weights
        self.verbose = verbose
        self._fitted = False
    
    def fit(self, X: Matrix, y: Vector) -> 'VotingRegressor':
        if self.weights and len(self.weights) != len(self.estimators):
            raise ValueError(f"Size mismatch between 'weights' and 'estimators'!")
        
        self.estimators_ = []
        for est in self.estimators:
            fitted_est = Clone(est).get
            if hasattr(fitted_est, 'verbose'):
                fitted_est.verbose = self.verbose
            
            fitted_est.fit(X, y)
            self.estimators_.append(fitted_est)
        
        self._fitted = True
        return self
    
    def predict(self, X: Matrix) -> Vector:
        if not self._fitted: raise NotFittedError(self)
        
        predictions = np.asarray([est.predict(X) for est in self.estimators_])
        avg_predictions = np.average(predictions, axis=0, weights=self.weights)
        return avg_predictions

    def score(self, X: Matrix, y: Vector, 
              metric: Evaluator = MeanSquaredError) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)
    
    def set_params(self, 
                   estimators: List[Estimator] = None,
                   weights: Vector[Scalar] = None) -> None:
        if estimators is not None: self.estimators = estimators
        if weights is not None: self.weights = weights

    def __getitem__(self, index: int) -> Estimator:
        return self.estimators_[index]




import numpy as np

from luma.interface.util import Matrix
from luma.interface.super import Transformer
from luma.interface.exception import NotFittedError


__all__ = (
    'StandardScaler', 
    'MinMaxScaler'
)


class StandardScaler(Transformer, Transformer.Feature):
    
    """
    Standard scaling is a data preprocessing technique to transform 
    the data so that it has a mean of 0 and a standard deviation of 1.
    """
    
    def __init__(self) -> None:
        self.mean = None
        self.std = None
        self._fitted = False

    def fit(self, X: Matrix) -> 'StandardScaler':
        self.mean = np.mean(X, axis=0)
        self.std = np.std(X, axis=0)
        self._fitted = True
        return self

    def transform(self, X: Matrix) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        return (X - self.mean) / self.std

    def fit_transform(self, X: Matrix) -> Matrix:
        self.fit(X)
        return self.transform(X)

    def inverse_transform(self, X: Matrix) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        return (X * self.std) + self.mean
    
    def set_params(self) -> None: ...


class MinMaxScaler(Transformer, Transformer.Feature):
    
    """
    MinMax scaling (also known as Min-Max normalization)
    to rescale features to a specific range, typically between 0 and 1. 
    The purpose of MinMax scaling is to transform the features in a way 
    that they fall within a specific interval.
    
    Parameters
    ----------
    `feature_range` : Range to be scaled
    
    """
    
    def __init__(self, feature_range: tuple = (0, 1)) -> None:
        self.feature_range = feature_range
        self.min = None
        self.max = None
        self._fitted = False

    def fit(self, X: Matrix) -> 'MinMaxScaler':
        self.min = np.min(X, axis=0)
        self.max = np.max(X, axis=0)
        self._fitted = True
        return self

    def transform(self, X: Matrix) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        min_val, max_val = self.feature_range
        scaled = (X - self.min) / (self.max - self.min) * (max_val - min_val)
        return scaled + min_val

    def fit_transform(self, X: Matrix) -> Matrix:
        self.fit(X)
        return self.transform(X)

    def inverse_transform(self, X: Matrix) -> Matrix:
        if not self._fitted: raise NotFittedError(self)
        min_val, max_val = self.feature_range
        original = (X - min_val) / (max_val - min_val) * (self.max - self.min)
        return original + self.min
    
    def set_params(self, feature_range: tuple = None) -> None:
        if feature_range is not None: self.feature_range = tuple(feature_range)




from luma.core.main import LUMA

from luma.interface.exception import NotFittedError, NotConvergedError
from luma.interface.exception import UnsupportedParameterError, ModelExtensionError
from luma.interface.super import Estimator, Transformer, Evaluator, Visualizer
from luma.interface.super import Supervised, Unsupervised, Distance
from luma.interface.util import Matrix, Vector, Scalar, TreeNode, NearestNeighbors
from luma.interface.util import SilhouetteUtil, DBUtil, KernelUtil
from luma.interface.util import Clone

from luma.classifier.logistic import LogisticRegressor, SoftmaxRegressor
from luma.classifier.naive_bayes import GaussianNaiveBayes, BernoulliNaiveBayes
from luma.classifier.svm import SVC, KernelSVC
from luma.classifier.tree import DecisionTreeClassifier
from luma.classifier.neighbors import KNNClassifier, AdaptiveKNNClassifier, WeightedKNNClassifier

from luma.clustering.kmeans import KMeansClustering, KMeansClusteringPlus, KMediansClustering
from luma.clustering.kmeans import KMedoidsClustering, MiniBatchKMeansClustering
from luma.clustering.kmeans import FuzzyKMeansClustering
from luma.clustering.hierarchy import AgglomerativeClustering, DivisiveClustering
from luma.clustering.spectral import SpectralClustering, NormalizedSpectralClustering
from luma.clustering.spectral import HierarchicalSpectralClustering, AdaptiveSpectralClustering
from luma.clustering.density import DBSCAN, OPTICS, DENCLUE, MeanShiftClustering
from luma.clustering.affinity import AffinityPropagation, AdaptiveAffinityPropagation
from luma.clustering.affinity import KernelAffinityPropagation
from luma.clustering.mixture import GaussianMixture, MultinomialMixture

from luma.ensemble.forest import RandomForestClassifier, RandomForestRegressor
from luma.ensemble.vote import VotingClassifier, VotingRegressor
from luma.ensemble.bagging import BaggingClassifier, BaggingRegressor
from luma.ensemble.boost import AdaBoostClassifier, AdaBoostRegressor

from luma.neural.single import PerceptronClassifier, PerceptronRegressor

from luma.metric.classification import Accuracy, Precision, Recall, F1Score, Specificity
from luma.metric.regression import MeanAbsoluteError, MeanSquaredError, RootMeanSquaredError
from luma.metric.regression import MeanAbsolutePercentageError, RSquaredScore
from luma.metric.clustering import SilhouetteCoefficient, DaviesBouldin
from luma.metric.distance import Euclidean, Manhattan, Chebyshev, Minkowski
from luma.metric.distance import CosineSimilarity, Correlation, Mahalanobis

from luma.model_selection.split import TrainTestSplit
from luma.model_selection.search import GridSearchCV
from luma.model_selection.cv import CrossValidator
from luma.model_selection.robust import RANSAC, MLESAC

from luma.preprocessing.scaler import StandardScaler, MinMaxScaler
from luma.preprocessing.encoder import OneHotEncoder, LabelEncoder, OrdinalEncoder
from luma.preprocessing.encoder import LabelBinarizer
from luma.preprocessing.imputer import SimpleImputer, KNNImputer, HotDeckImputer
from luma.preprocessing.outlier import LocalOutlierFactor

from luma.reduction.linear import PCA, LDA, TruncatedSVD, FactorAnalysis
from luma.reduction.nonlinear import KernelPCA
from luma.reduction.manifold import TSNE, SammonMapping, LaplacianEigenmap
from luma.reduction.manifold import MDS, MetricMDS, LandmarkMDS
from luma.reduction.manifold import LLE, ModifiedLLE, HessianLLE, LTSA
from luma.reduction.manifold import Isomap, ConformalIsomap
from luma.reduction.select import SBS, SFS, RFE

from luma.regressor.linear import LinearRegressor
from luma.regressor.linear import LassoRegressor, RidgeRegressor, ElasticNetRegressor
from luma.regressor.poly import PolynomialRegressor
from luma.regressor.general import PoissonRegressor
from luma.regressor.general import NegativeBinomialRegressor
from luma.regressor.general import GammaRegressor
from luma.regressor.general import BetaRegressor
from luma.regressor.general import InverseGaussianRegressor
from luma.regressor.svm import SVR, KernelSVR
from luma.regressor.tree import DecisionTreeRegressor
from luma.regressor.neighbors import KNNRegressor, AdaptiveKNNRegressor, WeightedKNNRegressor

from luma.pipe.pipeline import Pipeline

from luma.visual.eda import CorrelationBar, CorrelationHeatMap, JointPlot, MissingProportion
from luma.visual.graph import GraphPlot
from luma.visual.result import DecisionRegion, ClusterPlot, ROCCurve, PrecisionRecallCurve
from luma.visual.result import ConfusionMatrix, ResidualPlot

from luma.migrate.port import ModelPorter


if __name__ == '__main__':
    
    # ------------------- [ luma.core ] ------------------------
    LUMA
    
    # ----------------- [ luma.interface ] ---------------------
    NotFittedError, NotConvergedError,
    UnsupportedParameterError, ModelExtensionError
    
    Estimator, Transformer, Evaluator, Visualizer,
    Supervised, Unsupervised, Distance
    
    Matrix, Vector, Scalar, TreeNode, NearestNeighbors,
    SilhouetteUtil, DBUtil, KernelUtil,
    Clone
    
    # ----------------- [ luma.classifier ] --------------------
    LogisticRegressor, SoftmaxRegressor
    
    GaussianNaiveBayes, BernoulliNaiveBayes
    
    SVC, KernelSVC
    
    DecisionTreeClassifier
    
    KNNClassifier, AdaptiveKNNClassifier, WeightedKNNClassifier
    
    # ----------------- [ luma.clustering ] --------------------
    KMeansClustering, KMeansClusteringPlus, KMediansClustering,
    KMedoidsClustering, MiniBatchKMeansClustering,
    FuzzyKMeansClustering
    
    AgglomerativeClustering, DivisiveClustering
    
    SpectralClustering, NormalizedSpectralClustering,
    HierarchicalSpectralClustering, AdaptiveSpectralClustering
    
    DBSCAN, OPTICS, DENCLUE, MeanShiftClustering
    
    AffinityPropagation, AdaptiveAffinityPropagation,
    KernelAffinityPropagation
    
    GaussianMixture, MultinomialMixture
    
    # ------------------ [ luma.ensemble ] ---------------------
    RandomForestClassifier, RandomForestRegressor
    
    VotingClassifier, VotingRegressor
    
    BaggingClassifier, BaggingRegressor
    
    AdaBoostClassifier, AdaBoostRegressor
    
    # ------------------- [ luma.neural ] ----------------------
    PerceptronClassifier, PerceptronRegressor
    
    # ------------------- [ luma.metric ] ----------------------
    Accuracy, Precision, Recall, F1Score, Specificity
    
    MeanAbsoluteError, MeanSquaredError, RootMeanSquaredError,
    MeanAbsolutePercentageError, RSquaredScore
    
    SilhouetteCoefficient, DaviesBouldin
    
    Euclidean, Manhattan, Chebyshev, Minkowski,
    CosineSimilarity, Correlation, Mahalanobis
    
    # --------------- [ luma.module_selection ] ----------------
    TrainTestSplit
    
    GridSearchCV
    
    CrossValidator
    
    RANSAC, MLESAC
    
    # ---------------- [ luma.preprocessing ] ------------------
    StandardScaler, MinMaxScaler
    
    OneHotEncoder, LabelEncoder, OrdinalEncoder, LabelBinarizer
    
    SimpleImputer, KNNImputer, HotDeckImputer
    
    LocalOutlierFactor
    
    # ----------------- [ luma.reduction ] ---------------------
    PCA, LDA, TruncatedSVD, FactorAnalysis
    
    KernelPCA
    
    TSNE, SammonMapping, LaplacianEigenmap,
    MDS, MetricMDS, LandmarkMDS,
    LLE, ModifiedLLE, HessianLLE, LTSA,
    Isomap, ConformalIsomap
    
    SBS, SFS, RFE
    
    # ----------------- [ luma.regressor ] ---------------------
    LinearRegressor,
    RidgeRegressor, LassoRegressor, ElasticNetRegressor
    
    PolynomialRegressor
    
    PoissonRegressor, NegativeBinomialRegressor, GammaRegressor,
    BetaRegressor, InverseGaussianRegressor
    
    SVR, KernelSVR
    
    DecisionTreeRegressor
    
    KNNRegressor, AdaptiveKNNRegressor, WeightedKNNRegressor
    
    # -------------------- [ luma.pipe ] -----------------------
    Pipeline
    
    # ------------------- [ luma.visual ] ----------------------
    CorrelationBar, CorrelationHeatMap, JointPlot,
    MissingProportion
    
    GraphPlot
    
    DecisionRegion, ClusterPlot, ROCCurve, PrecisionRecallCurve, 
    ConfusionMatrix, ResidualPlot
    
    # ------------------ [ luma.migrate ] ----------------------
    ModelPorter



from typing import List, Literal, Optional, Tuple
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap 
import numpy as np

from luma.interface.util import Matrix, Vector
from luma.interface.super import Visualizer, Estimator
from luma.preprocessing.encoder import LabelBinarizer
from luma.metric.classification import Precision, Recall, Specificity


__all__ = (
    'DecisionRegion', 
    'ClusterPlot',
    'ROCCurve',
    'PrecisionRecallCurve',
    'ConfusionMatrix',
    'ResidualPlot'
)


class DecisionRegion(Visualizer):
    def __init__(self, 
                 estimator: Estimator,
                 X: Matrix, 
                 y: Optional[Matrix] = None, 
                 title: str | Literal['auto'] = 'auto', 
                 xlabel: str = r'$x_1$', 
                 ylabel: str = r'$x_2$',
                 cmap: ListedColormap = 'rainbow',
                 alpha: float = 0.4) -> None:
        self.estimator = estimator
        self.X = X
        self.y = y
        self.title = title
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.cmap = cmap
        self.alpha = alpha
        
        if self.title == 'auto':
            self.title = type(self.estimator).__name__

        if self.y is None and hasattr(self.estimator, 'labels'):
            self.y = self.estimator.labels

    def plot(self, 
             ax: Optional[plt.Axes] = None, 
             size: float = 250, 
             scale: float = 10.0, 
             show: bool = False) -> plt.Axes:
        x1_min, x1_max = self.X[:, 0].min(), self.X[:, 0].max()
        x2_min, x2_max = self.X[:, 1].min(), self.X[:, 1].max()
        delta_1, delta_2 = (x1_max - x1_min) / size, (x2_max - x2_min) / size
        
        x1_min, x1_max = x1_min - delta_1 * scale, x1_max + delta_1 * scale
        x2_min, x2_max = x2_min - delta_2 * scale, x2_max + delta_2 * scale
        xx1, xx2 = np.meshgrid(np.arange(x1_min, x1_max, delta_1), 
                               np.arange(x2_min, x2_max, delta_2))
        
        Z = self.estimator.predict(np.array([xx1.ravel(), xx2.ravel()]).T)
        Z = Z.reshape(xx1.shape)
        
        if ax is None: _, ax = plt.subplots()
        ax.contourf(xx1, xx2, Z, 
                    alpha=self.alpha, 
                    cmap=self.cmap, 
                    levels=len(np.unique(self.y)))
        
        ax.set_xlim(xx1.min(), xx1.max())
        ax.set_ylim(xx2.min(), xx2.max())

        ax.scatter(self.X[:, 0], self.X[:, 1], 
                   c=self.y, 
                   cmap=self.cmap, 
                   alpha=0.8, 
                   edgecolors='black')
        
        ax.set_xlabel(self.xlabel)
        ax.set_ylabel(self.ylabel)
        ax.set_title(self.title)
        ax.figure.tight_layout()
        
        if show: plt.show()
        return ax


class ClusterPlot(Visualizer):
    def __init__(self,
                 estimator: Estimator,
                 X: Matrix,
                 title: str | Literal['auto'] = 'auto',
                 xlabel: str = r'$x_1$',
                 ylabel: str = r'$x_2$',
                 cmap: ListedColormap = 'rainbow',
                 alpha: float = 0.8) -> None:
        self.estimator = estimator
        self.X = X
        self.title = title
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.cmap = cmap
        self.alpha = alpha
        self.labels = self.estimator.labels
        
        if self.title == 'auto':
            self.title = type(self.estimator).__name__
    
    def plot(self, 
             ax: Optional[plt.Axes] = None, 
             show: bool = False) -> plt.Axes:
        if ax is None: _, ax = plt.subplots()
        ax.scatter(self.X[self.labels == -1, 0], 
                   self.X[self.labels == -1, 1],
                   marker='x',
                   c='black', 
                   label='Noise')
        
        ax.scatter(self.X[self.labels != -1, 0], 
                   self.X[self.labels != -1, 1], 
                   marker='o',
                   c=self.labels[self.labels != -1],
                   cmap=self.cmap,
                   alpha=self.alpha,
                   edgecolors='black')
        
        ax.set_title(self.title)
        ax.set_xlabel(self.xlabel)
        ax.set_ylabel(self.ylabel)
        
        if len(self.X[self.labels == -1]): ax.legend()
        ax.figure.tight_layout()

        if show: plt.show()
        return ax


class ROCCurve(Visualizer):
    def __init__(self,
                 y_true: Vector,
                 y_scores: Matrix) -> None:
        self.y_true = y_true
        self.y_scores = y_scores
        self.n_classes = y_scores.shape[1]
    
    def plot(self, 
             ax: Optional[plt.Axes] = None, 
             show: bool = False) -> plt.Axes:
        if ax is None: _, ax = plt.subplots()
        y_binary = LabelBinarizer().fit_transform(self.y_true)
        
        fprs, tprs = [], []
        for cl in range(self.n_classes):
            fpr, tpr = self._fpr_tpr(y_binary[:, cl], self.y_scores[:, cl])
            fprs.append(fpr)
            tprs.append(tpr)
            
            auc = self._auc(fpr, tpr)
            ax.plot(fpr, tpr, 
                    linewidth=2,
                    label=f'ROC Curve {cl} (area = {auc:.2f})')
        
        mean_fpr, mean_tpr = np.mean(fprs, axis=0), np.mean(tprs, axis=0)
        mean_auc = self._auc(mean_fpr, mean_tpr)
        
        ax.plot(mean_fpr, mean_tpr,
                color='dimgray', 
                linewidth=2,
                linestyle='--', 
                label=f'Mean ROC (area = {mean_auc:.2f})')
        
        ax.plot([0, 1], [0, 1], 
                color='darkgray', 
                linestyle=':', 
                label='Random Guessing')
        
        ax.set_xlim([-0.05, 1.05])
        ax.set_ylim([-0.05, 1.05])
        ax.set_xlabel('False Positive Rate')
        
        ax.set_ylabel('True Positive Rate')
        ax.set_title('ROC Curve')
        ax.legend(loc="lower right")
        ax.figure.tight_layout()

        if show: plt.show()
        return ax
    
    def _fpr_tpr(self, y_true: Vector, y_score: Vector) -> Tuple[Vector, Vector]:
        thresholds = np.sort(y_score)[::-1]
        fpr, tpr = [], []
        for threshold in thresholds:
            y_pred = y_score >= threshold
            tpr.append(Recall.score(y_true, y_pred))
            fpr.append(1 - Specificity.score(y_true, y_pred))
        
        return np.array(fpr), np.array(tpr)
    
    def _auc(self, fpr: Vector, tpr: Vector) -> float:
        auc = 0
        for i in range(1, len(fpr)):
            auc += (fpr[i] - fpr[i - 1]) * (tpr[i] + tpr[i - 1]) / 2
        return auc


class PrecisionRecallCurve(Visualizer):
    def __init__(self,
                 y_true: Vector,
                 y_scores: Matrix) -> None:
        self.y_true = y_true
        self.y_scores = y_scores
        self.n_classes = y_scores.shape[1]

    def plot(self,
             ax: Optional[plt.Axes] = None,
             show: bool = False) -> plt.Axes:
        if ax is None: _, ax = plt.subplots()
        y_binary = LabelBinarizer().fit_transform(self.y_true)

        pres, recs = [], []
        for cl in range(self.n_classes):
            pre, rec = self._pre_rec(y_binary[:, cl], self.y_scores[:, cl])
            pres.append(pre)
            recs.append(rec)

            ap = self._average_precision(pre, rec)
            ax.plot(rec, pre,
                    linewidth=2,
                    label=f'PR Curve {cl} (AP = {ap:.2f})')

        mean_pre, mean_rec = np.mean(pres, axis=0), np.mean(recs, axis=0)
        mean_ap = self._average_precision(mean_pre, mean_rec)

        ax.plot(mean_rec, mean_pre,
                color='dimgray',
                linewidth=2,
                linestyle='--',
                label=f'Mean PR (AP = {mean_ap:.2f})')
        
        ax.plot([1, 0], [0, 1], 
                color='darkgray', 
                linestyle=':', 
                label='Random Guessing')

        ax.set_xlim([-0.05, 1.05])
        ax.set_ylim([-0.05, 1.05])
        ax.set_xlabel('Recall')
        ax.set_ylabel('Precision')
        ax.set_title('Precision-Recall Curve')
        ax.legend(loc="lower left")
        ax.figure.tight_layout()

        if show: plt.show()
        return ax
    
    def _pre_rec(self, y_true: Vector, y_score: Vector) -> Tuple[Vector, Vector]:
        thresholds = np.sort(y_score)[::-1]
        pre, rec = [], []
        for threshold in thresholds:
            y_pred = y_score >= threshold
            pre.append(Precision.score(y_true, y_pred))
            rec.append(Recall.score(y_true, y_pred))

        return np.array(pre), np.array(rec)
    
    def _average_precision(self, pre: Vector, rec: Vector) -> float:
        ap = 0
        for i in range(1, len(rec)):
            ap += (rec[i] - rec[i - 1]) * pre[i]
        return ap


class ConfusionMatrix(Visualizer):
    def __init__(self, 
                 y_true: Vector, 
                 y_pred: Vector, 
                 labels: List[str] | None = None,
                 cmap: ListedColormap = 'Blues') -> None:
        self.cmap = cmap
        self.conf_matrix = self._confusion_matrix(y_true, y_pred)
        
        if labels is None:
            self.labels = np.arange(len(np.unique(y_true)))

    def plot(self, ax: Optional[plt.Axes] = None, show: bool = False) -> plt.Axes:
        if ax is None: ax = plt.gca()
        
        cax = ax.imshow(self.conf_matrix, interpolation='nearest', cmap=self.cmap)
        ax.set_title('Confusion Matrix')
        plt.colorbar(cax, ax=ax)

        tick_marks = np.arange(len(self.labels))
        ax.set_xticks(tick_marks)
        ax.set_xticklabels(self.labels, rotation=45)
        ax.set_yticks(tick_marks)
        ax.set_yticklabels(self.labels)

        thresh = self.conf_matrix.max() / 2.
        for i, j in np.ndindex(self.conf_matrix.shape):
            ax.text(j, i, format(self.conf_matrix[i, j], 'd'),
                    horizontalalignment="center",
                    color="white" if self.conf_matrix[i, j] > thresh else "black")

        ax.set_ylabel('True label')
        ax.set_xlabel('Predicted label')
        ax.set_aspect('equal', adjustable='box')

        if show: plt.show()
        return ax
    
    def _confusion_matrix(self, y_true: Vector, y_pred: Vector) -> Matrix:
        unique_labels = np.unique(np.concatenate((y_true, y_pred)))
        matrix = np.zeros((len(unique_labels), len(unique_labels)), dtype=int)
        for true, pred in zip(y_true, y_pred):
            matrix[true, pred] += 1
        
        return matrix


class ResidualPlot(Visualizer):
    def __init__(self, 
                 estimator: Estimator,
                 X: Matrix,
                 y: Vector,
                 alpha: float = 0.7,
                 cmap: ListedColormap = 'RdYlBu') -> None:
        self.estimator = estimator
        self.X = X
        self.y = y
        self.alpha = alpha
        self.cmap = cmap
    
    def plot(self, 
             ax: Optional[plt.Axes] = None,
             show: bool = False) -> plt.Axes:
        if ax is None: _, ax = plt.subplots()
        resid = self._calculate_residuals()
        
        ax.scatter(self.X, resid, 
                   c=resid, 
                   s=20, 
                   cmap=self.cmap, 
                   alpha=self.alpha)

        ax.axhline(y=0, c='black', 
                   lw=2, label='Perfect Fit')
        
        ax.axhline(y=resid.mean(), c='gray', ls='--', 
                   lw=2, label='Average Residual')
        
        ax.set_xlabel('Predicted Values')
        ax.set_ylabel('Residuals')
        ax.set_title(f'Residual Plot of {type(self.estimator).__name__}')
        ax.legend()
        ax.figure.tight_layout()

        if show: plt.show()
        return ax

    def _calculate_residuals(self) -> Vector:
        predictions = self.estimator.predict(self.X)
        residuals = self.y - predictions
        return residuals

